// code for Cray Aries DMAPP
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "pmi.h"
#include "dmapp.h"

#define NUM_ITERATIONS 10000

/* IR code:
process 0
shared X = 0, Y = 1;
local a, b;
X = get(atm,atm)(Y,0);
Y = get(atm,atm)(Y,0);
load(atm) a = X;
load(atm) b = Y;
// ---

output((a == 1) && (b == 1));
output((a == 0) && (b == 1));
*/

int main(int argc, char **argv) {
    char            *expl = NULL;
    int             pe = -1;
    int             npes = -1;
    int             target_pe = -1;

    long                a = -1L;
    long                b = -1L;
    long                cnt_0 = 0;
    long                cnt_1 = 0;
    long                *X = NULL;
    long                *Y = NULL;
    long                *a_sender = NULL;
    long                *a_receiver = NULL;
    long                *a_global = NULL;
    long                *b_sender = NULL;
    long                *b_receiver = NULL;
    long                *b_global = NULL;
    dmapp_return_t      status;
    dmapp_rma_attrs_t   actual_args = {0}, rma_args = {0};
    dmapp_jobinfo_t     job;
    dmapp_seg_desc_t    *seg = NULL;

    /* Set the RMA parameters. */
    rma_args.put_relaxed_ordering = DMAPP_ROUTING_DETERMINISTIC;
    rma_args.max_outstanding_nb = DMAPP_DEF_OUTSTANDING_NB;
    rma_args.offload_threshold = DMAPP_OFFLOAD_THRESHOLD;
    rma_args.max_concurrency = 1;

    /* Initialize DMAPP resources before executing any other DMAPP calls. */
    status = dmapp_init(&rma_args, &actual_args);
    if(status != DMAPP_RC_SUCCESS) {
        dmapp_explain_error(status, (const char **)&expl);
        fprintf(stderr, " dmapp_init returned %d - %s - %s \n", status, dmapp_err_str[status], expl);
        exit(1);
    }

    /* Retrieve information about job details such as PE id and number of PEs. */
    status = dmapp_get_jobinfo(&job);
    if (status != DMAPP_RC_SUCCESS) {
        dmapp_explain_error(status, (const char **)&expl);
        fprintf(stderr, " dmapp_get_jobinfo FAILED: w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
        exit(1);
    }
    pe = job.pe;
    npes = job.npes;
    target_pe = npes - pe - 1;



    /* Specify in which segment the remote memory region (the source) lies.
       In this case, it is the heap (see above). */
    seg = &(job.sheap_seg);

    for(int i = 0; i < NUM_ITERATIONS; i++) {

        /* Allocate remotely accessible memory for the source and target buffers.
           Only memory in the data segment or the sheap is remotely accessible.
           Here we allocate from the sheap. */
	X = (long *)dmapp_sheap_malloc(sizeof(long));
	if (X == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)X);
		exit(1);
	}
	 *X = 0;
	Y = (long *)dmapp_sheap_malloc(sizeof(long));
	if (Y == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)Y);
		exit(1);
	}
	 *Y = 1;
	a_sender = (long *)dmapp_sheap_malloc(sizeof(long));
	if (a_sender == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)a_sender);
		exit(1);
	}
	a_receiver = (long *)dmapp_sheap_malloc(sizeof(long));
	if (a_receiver == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)a_receiver);
		exit(1);
	}
	a_global = (long *)dmapp_sheap_malloc(sizeof(long));
	if (a_global == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)a_global);
		exit(1);
	}
	b_sender = (long *)dmapp_sheap_malloc(sizeof(long));
	if (b_sender == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)b_sender);
		exit(1);
	}
	b_receiver = (long *)dmapp_sheap_malloc(sizeof(long));
	if (b_receiver == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)b_receiver);
		exit(1);
	}
	b_global = (long *)dmapp_sheap_malloc(sizeof(long));
	if (b_global == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)b_global);
		exit(1);
	}

        /* Synchronize to make sure everyone's buffers are initialized before data transfer is started.  */
        PMI_Barrier();

        /* body */
	if (rand() % 2 == 1) { usleep(1); }
	status = dmapp_get_nbi(X, Y, seg, 0, 1, DMAPP_QW);
	if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
	if (rand() % 2 == 1) { usleep(1); }
	status = dmapp_get_nbi(Y, Y, seg, 0, 1, DMAPP_QW);
	if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
	if (rand() % 2 == 1) { usleep(1); }
	a = *X;
	if (rand() % 2 == 1) { usleep(1); }
	b = *Y;

        /* Synchronize before verifying the data. */
        dmapp_gsync_wait();
        PMI_Barrier();

        /* Verify the data. */
	if (1) {
		*a_receiver = a;
		*b_receiver = b;
	}
	dmapp_gsync_wait();
	PMI_Barrier();

	/* Load results from global _received variables to locals */
	if (1) {
		a = *a_receiver;
		b = *b_receiver;
	}

	/* Check which output is observed */
	if (1) {
		if ((a == 1) && (b == 1)) cnt_0++;
		else if ((a == 0) && (b == 1)) cnt_1++;
		else fprintf(stderr, "UNEXPECTED: a=%lld b=%lld\n", a, b);
	}

        /* Free buffers allocated from sheap. */
	dmapp_sheap_free(X);
	dmapp_sheap_free(Y);
	dmapp_sheap_free(a_sender);
	dmapp_sheap_free(a_receiver);
	dmapp_sheap_free(a_global);
	dmapp_sheap_free(b_sender);
	dmapp_sheap_free(b_receiver);
	dmapp_sheap_free(b_global);
    }

    /* Print output stats */
	if (1) {
		fprintf(stderr, "OBSERVED ((a == 1) && (b == 1)): %ld\n", cnt_0);
		fprintf(stderr, "OBSERVED ((a == 0) && (b == 1)): %ld\n", cnt_1);
	}

    /* Release DMAPP resources. This is a mandatory call. */
    status = dmapp_finalize();
    if (status != DMAPP_RC_SUCCESS) {
        dmapp_explain_error(status, (const char **)&expl);
        fprintf(stderr, " dmapp_finalize FAILED w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
        exit(1);
    }

    return(0);
}

