// code for Cray Aries DMAPP
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "pmi.h"
#include "dmapp.h"

#define NUM_ITERATIONS 10000

/* IR code:
process 0
shared X = 0;
local a;
store(atm) X = 3;
load(atm) a = X;

process 1
shared Y = 1;
local b, c, d;
store(atm) Y = 2;
load(atm) d = Y;
store(atm) Y = 4;
load(atm) b = Y;
load(atm) c = Y;
// ---

output((d == 2) && (c == 4) && (a == 3) && (b == 4));
*/

int main(int argc, char **argv) {
    char            *expl = NULL;
    int             pe = -1;
    int             npes = -1;
    int             target_pe = -1;

    long                a = -1L;
    long                b = -1L;
    long                c = -1L;
    long                d = -1L;
    long                cnt_0 = 0;
    long                *X = NULL;
    long                *a_sender = NULL;
    long                *a_receiver = NULL;
    long                *a_global = NULL;
    long                *Y = NULL;
    long                *b_sender = NULL;
    long                *b_receiver = NULL;
    long                *b_global = NULL;
    long                *c_sender = NULL;
    long                *c_receiver = NULL;
    long                *c_global = NULL;
    long                *d_sender = NULL;
    long                *d_receiver = NULL;
    long                *d_global = NULL;
    dmapp_return_t      status;
    dmapp_rma_attrs_t   actual_args = {0}, rma_args = {0};
    dmapp_jobinfo_t     job;
    dmapp_seg_desc_t    *seg = NULL;

    /* Set the RMA parameters. */
    rma_args.put_relaxed_ordering = DMAPP_ROUTING_DETERMINISTIC;
    rma_args.max_outstanding_nb = DMAPP_DEF_OUTSTANDING_NB;
    rma_args.offload_threshold = DMAPP_OFFLOAD_THRESHOLD;
    rma_args.max_concurrency = 1;

    /* Initialize DMAPP resources before executing any other DMAPP calls. */
    status = dmapp_init(&rma_args, &actual_args);
    if(status != DMAPP_RC_SUCCESS) {
        dmapp_explain_error(status, (const char **)&expl);
        fprintf(stderr, " dmapp_init returned %d - %s - %s \n", status, dmapp_err_str[status], expl);
        exit(1);
    }

    /* Retrieve information about job details such as PE id and number of PEs. */
    status = dmapp_get_jobinfo(&job);
    if (status != DMAPP_RC_SUCCESS) {
        dmapp_explain_error(status, (const char **)&expl);
        fprintf(stderr, " dmapp_get_jobinfo FAILED: w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
        exit(1);
    }
    pe = job.pe;
    npes = job.npes;
    target_pe = npes - pe - 1;



    /* Specify in which segment the remote memory region (the source) lies.
       In this case, it is the heap (see above). */
    seg = &(job.sheap_seg);

    for(int i = 0; i < NUM_ITERATIONS; i++) {

        /* Allocate remotely accessible memory for the source and target buffers.
           Only memory in the data segment or the sheap is remotely accessible.
           Here we allocate from the sheap. */
	X = (long *)dmapp_sheap_malloc(sizeof(long));
	if (X == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)X);
		exit(1);
	}
	if ((pe % 2) == 0)  *X = 0;
	a_sender = (long *)dmapp_sheap_malloc(sizeof(long));
	if (a_sender == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)a_sender);
		exit(1);
	}
	a_receiver = (long *)dmapp_sheap_malloc(sizeof(long));
	if (a_receiver == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)a_receiver);
		exit(1);
	}
	a_global = (long *)dmapp_sheap_malloc(sizeof(long));
	if (a_global == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)a_global);
		exit(1);
	}
	Y = (long *)dmapp_sheap_malloc(sizeof(long));
	if (Y == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)Y);
		exit(1);
	}
	if ((pe % 2) == 1)  *Y = 1;
	b_sender = (long *)dmapp_sheap_malloc(sizeof(long));
	if (b_sender == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)b_sender);
		exit(1);
	}
	b_receiver = (long *)dmapp_sheap_malloc(sizeof(long));
	if (b_receiver == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)b_receiver);
		exit(1);
	}
	b_global = (long *)dmapp_sheap_malloc(sizeof(long));
	if (b_global == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)b_global);
		exit(1);
	}
	c_sender = (long *)dmapp_sheap_malloc(sizeof(long));
	if (c_sender == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)c_sender);
		exit(1);
	}
	c_receiver = (long *)dmapp_sheap_malloc(sizeof(long));
	if (c_receiver == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)c_receiver);
		exit(1);
	}
	c_global = (long *)dmapp_sheap_malloc(sizeof(long));
	if (c_global == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)c_global);
		exit(1);
	}
	d_sender = (long *)dmapp_sheap_malloc(sizeof(long));
	if (d_sender == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)d_sender);
		exit(1);
	}
	d_receiver = (long *)dmapp_sheap_malloc(sizeof(long));
	if (d_receiver == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)d_receiver);
		exit(1);
	}
	d_global = (long *)dmapp_sheap_malloc(sizeof(long));
	if (d_global == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)d_global);
		exit(1);
	}

        /* Synchronize to make sure everyone's buffers are initialized before data transfer is started.  */
        PMI_Barrier();

        /* body */
	if ((pe % 2) == 0) {
		if (rand() % 2 == 1) { usleep(5 + (rand()%10)); }
		if (rand() % 2 == 1) { usleep(1); }
		*X = 3;
		if (rand() % 2 == 1) { usleep(1); }
		a = *X;
	}
	if ((pe % 2) == 1) {
		if (rand() % 2 == 1) { usleep(1); }
		*Y = 2;
		if (rand() % 2 == 1) { usleep(1); }
		d = *Y;
		if (rand() % 2 == 1) { usleep(1); }
		*Y = 4;
		if (rand() % 2 == 1) { usleep(1); }
		b = *Y;
		if (rand() % 2 == 1) { usleep(1); }
		c = *Y;
	}

        /* Synchronize before verifying the data. */
        dmapp_gsync_wait();
        PMI_Barrier();

        /* Verify the data. */
	if ((pe % 2) == 0) {
		*a_receiver = a;
	}
	if ((pe % 2) == 1) {
		*b_sender = b;
		status = dmapp_put_nbi(b_receiver, seg, 0, b_sender, 1, DMAPP_QW);
		if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
		*c_sender = c;
		status = dmapp_put_nbi(c_receiver, seg, 0, c_sender, 1, DMAPP_QW);
		if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
		*d_sender = d;
		status = dmapp_put_nbi(d_receiver, seg, 0, d_sender, 1, DMAPP_QW);
		if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
	}
	dmapp_gsync_wait();
	PMI_Barrier();

	/* Load results from global _received variables to locals */
	if ((pe % 2) == 0) {
		a = *a_receiver;
		b = *b_receiver;
		c = *c_receiver;
		d = *d_receiver;
	}

	/* Check which output is observed */
	if ((pe % 2) == 0) {
		if ((((d == 2) && (c == 4)) && (a == 3)) && (b == 4)) cnt_0++;
		else fprintf(stderr, "UNEXPECTED: a=%lld b=%lld c=%lld d=%lld\n", a, b, c, d);
	}

        /* Free buffers allocated from sheap. */
	dmapp_sheap_free(X);
	dmapp_sheap_free(a_sender);
	dmapp_sheap_free(a_receiver);
	dmapp_sheap_free(a_global);
	dmapp_sheap_free(Y);
	dmapp_sheap_free(b_sender);
	dmapp_sheap_free(b_receiver);
	dmapp_sheap_free(b_global);
	dmapp_sheap_free(c_sender);
	dmapp_sheap_free(c_receiver);
	dmapp_sheap_free(c_global);
	dmapp_sheap_free(d_sender);
	dmapp_sheap_free(d_receiver);
	dmapp_sheap_free(d_global);
    }

    /* Print output stats */
	if ((pe % 2) == 0) {
		fprintf(stderr, "OBSERVED ((((d == 2) && (c == 4)) && (a == 3)) && (b == 4)): %ld\n", cnt_0);
	}

    /* Release DMAPP resources. This is a mandatory call. */
    status = dmapp_finalize();
    if (status != DMAPP_RC_SUCCESS) {
        dmapp_explain_error(status, (const char **)&expl);
        fprintf(stderr, " dmapp_finalize FAILED w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
        exit(1);
    }

    return(0);
}

