// code for Cray Aries DMAPP
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "pmi.h"
#include "dmapp.h"

#define NUM_ITERATIONS 10000

/* IR code:
process 0
shared X = 1;
local a, b, c;
X = get(atm,atm)(Y,1);
load(atm) b = X;
X = get(atm,atm)(Z,1);
load(atm) a = X;
load(atm) c = X;

process 1
shared Z = 2, Y = 0;
// ---

output((c == 2) && (a == 2) && (b == 1));
output((c == 0) && (a == 0) && (b == 0));
output((c == 2) && (a == 2) && (b == 0));
output((c == 2) && (a == 0) && (b == 0));
output((c == 1) && (a == 1) && (b == 1));
output((c == 0) && (a == 2) && (b == 1));
output((c == 0) && (a == 0) && (b == 1));
output((c == 0) && (a == 1) && (b == 1));
output((c == 2) && (a == 0) && (b == 1));
output((c == 2) && (a == 1) && (b == 1));
*/

int main(int argc, char **argv) {
    char            *expl = NULL;
    int             pe = -1;
    int             npes = -1;
    int             target_pe = -1;

    long                a = -1L;
    long                b = -1L;
    long                c = -1L;
    long                cnt_0 = 0;
    long                cnt_1 = 0;
    long                cnt_2 = 0;
    long                cnt_3 = 0;
    long                cnt_4 = 0;
    long                cnt_5 = 0;
    long                cnt_6 = 0;
    long                cnt_7 = 0;
    long                cnt_8 = 0;
    long                cnt_9 = 0;
    long                *X = NULL;
    long                *a_sender = NULL;
    long                *a_receiver = NULL;
    long                *a_global = NULL;
    long                *b_sender = NULL;
    long                *b_receiver = NULL;
    long                *b_global = NULL;
    long                *c_sender = NULL;
    long                *c_receiver = NULL;
    long                *c_global = NULL;
    long                *Z = NULL;
    long                *Y = NULL;
    dmapp_return_t      status;
    dmapp_rma_attrs_t   actual_args = {0}, rma_args = {0};
    dmapp_jobinfo_t     job;
    dmapp_seg_desc_t    *seg = NULL;

    /* Set the RMA parameters. */
    rma_args.put_relaxed_ordering = DMAPP_ROUTING_DETERMINISTIC;
    rma_args.max_outstanding_nb = DMAPP_DEF_OUTSTANDING_NB;
    rma_args.offload_threshold = DMAPP_OFFLOAD_THRESHOLD;
    rma_args.max_concurrency = 1;

    /* Initialize DMAPP resources before executing any other DMAPP calls. */
    status = dmapp_init(&rma_args, &actual_args);
    if(status != DMAPP_RC_SUCCESS) {
        dmapp_explain_error(status, (const char **)&expl);
        fprintf(stderr, " dmapp_init returned %d - %s - %s \n", status, dmapp_err_str[status], expl);
        exit(1);
    }

    /* Retrieve information about job details such as PE id and number of PEs. */
    status = dmapp_get_jobinfo(&job);
    if (status != DMAPP_RC_SUCCESS) {
        dmapp_explain_error(status, (const char **)&expl);
        fprintf(stderr, " dmapp_get_jobinfo FAILED: w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
        exit(1);
    }
    pe = job.pe;
    npes = job.npes;
    target_pe = npes - pe - 1;



    /* Specify in which segment the remote memory region (the source) lies.
       In this case, it is the heap (see above). */
    seg = &(job.sheap_seg);

    for(int i = 0; i < NUM_ITERATIONS; i++) {

        /* Allocate remotely accessible memory for the source and target buffers.
           Only memory in the data segment or the sheap is remotely accessible.
           Here we allocate from the sheap. */
	X = (long *)dmapp_sheap_malloc(sizeof(long));
	if (X == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)X);
		exit(1);
	}
	if ((pe % 2) == 0)  *X = 1;
	a_sender = (long *)dmapp_sheap_malloc(sizeof(long));
	if (a_sender == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)a_sender);
		exit(1);
	}
	a_receiver = (long *)dmapp_sheap_malloc(sizeof(long));
	if (a_receiver == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)a_receiver);
		exit(1);
	}
	a_global = (long *)dmapp_sheap_malloc(sizeof(long));
	if (a_global == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)a_global);
		exit(1);
	}
	b_sender = (long *)dmapp_sheap_malloc(sizeof(long));
	if (b_sender == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)b_sender);
		exit(1);
	}
	b_receiver = (long *)dmapp_sheap_malloc(sizeof(long));
	if (b_receiver == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)b_receiver);
		exit(1);
	}
	b_global = (long *)dmapp_sheap_malloc(sizeof(long));
	if (b_global == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)b_global);
		exit(1);
	}
	c_sender = (long *)dmapp_sheap_malloc(sizeof(long));
	if (c_sender == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)c_sender);
		exit(1);
	}
	c_receiver = (long *)dmapp_sheap_malloc(sizeof(long));
	if (c_receiver == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)c_receiver);
		exit(1);
	}
	c_global = (long *)dmapp_sheap_malloc(sizeof(long));
	if (c_global == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)c_global);
		exit(1);
	}
	Z = (long *)dmapp_sheap_malloc(sizeof(long));
	if (Z == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)Z);
		exit(1);
	}
	if ((pe % 2) == 1)  *Z = 2;
	Y = (long *)dmapp_sheap_malloc(sizeof(long));
	if (Y == NULL) {
		fprintf(stderr, " sheap_malloc'd failed ? 0x%lx\n", (long)Y);
		exit(1);
	}
	if ((pe % 2) == 1)  *Y = 0;

        /* Synchronize to make sure everyone's buffers are initialized before data transfer is started.  */
        PMI_Barrier();

        /* body */
	if ((pe % 2) == 0) {
		if (rand() % 2 == 1) { usleep(5 + (rand()%10)); }
		if (rand() % 2 == 1) { usleep(1); }
		status = dmapp_get_nbi(X, Y, seg, 1, 1, DMAPP_QW);
			if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
		if (rand() % 2 == 1) { usleep(1); }
		b = *X;
		if (rand() % 2 == 1) { usleep(1); }
		status = dmapp_get_nbi(X, Z, seg, 1, 1, DMAPP_QW);
			if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
		if (rand() % 2 == 1) { usleep(1); }
		a = *X;
		if (rand() % 2 == 1) { usleep(1); }
		c = *X;
	}
	if ((pe % 2) == 1) {
	}

        /* Synchronize before verifying the data. */
        dmapp_gsync_wait();
        PMI_Barrier();

        /* Verify the data. */
	if ((pe % 2) == 0) {
		*a_receiver = a;
		*b_receiver = b;
		*c_receiver = c;
	}
	dmapp_gsync_wait();
	PMI_Barrier();

	/* Load results from global _received variables to locals */
	if ((pe % 2) == 0) {
		a = *a_receiver;
		b = *b_receiver;
		c = *c_receiver;
	}

	/* Check which output is observed */
	if ((pe % 2) == 0) {
		if (((c == 2) && (a == 2)) && (b == 1)) cnt_0++;
		else if (((c == 0) && (a == 0)) && (b == 0)) cnt_1++;
		else if (((c == 2) && (a == 2)) && (b == 0)) cnt_2++;
		else if (((c == 2) && (a == 0)) && (b == 0)) cnt_3++;
		else if (((c == 1) && (a == 1)) && (b == 1)) cnt_4++;
		else if (((c == 0) && (a == 2)) && (b == 1)) cnt_5++;
		else if (((c == 0) && (a == 0)) && (b == 1)) cnt_6++;
		else if (((c == 0) && (a == 1)) && (b == 1)) cnt_7++;
		else if (((c == 2) && (a == 0)) && (b == 1)) cnt_8++;
		else if (((c == 2) && (a == 1)) && (b == 1)) cnt_9++;
		else fprintf(stderr, "UNEXPECTED: a=%lld b=%lld c=%lld\n", a, b, c);
	}

        /* Free buffers allocated from sheap. */
	dmapp_sheap_free(X);
	dmapp_sheap_free(a_sender);
	dmapp_sheap_free(a_receiver);
	dmapp_sheap_free(a_global);
	dmapp_sheap_free(b_sender);
	dmapp_sheap_free(b_receiver);
	dmapp_sheap_free(b_global);
	dmapp_sheap_free(c_sender);
	dmapp_sheap_free(c_receiver);
	dmapp_sheap_free(c_global);
	dmapp_sheap_free(Z);
	dmapp_sheap_free(Y);
    }

    /* Print output stats */
	if ((pe % 2) == 0) {
		fprintf(stderr, "OBSERVED (((c == 2) && (a == 2)) && (b == 1)): %ld\n", cnt_0);
		fprintf(stderr, "OBSERVED (((c == 0) && (a == 0)) && (b == 0)): %ld\n", cnt_1);
		fprintf(stderr, "OBSERVED (((c == 2) && (a == 2)) && (b == 0)): %ld\n", cnt_2);
		fprintf(stderr, "OBSERVED (((c == 2) && (a == 0)) && (b == 0)): %ld\n", cnt_3);
		fprintf(stderr, "OBSERVED (((c == 1) && (a == 1)) && (b == 1)): %ld\n", cnt_4);
		fprintf(stderr, "OBSERVED (((c == 0) && (a == 2)) && (b == 1)): %ld\n", cnt_5);
		fprintf(stderr, "OBSERVED (((c == 0) && (a == 0)) && (b == 1)): %ld\n", cnt_6);
		fprintf(stderr, "OBSERVED (((c == 0) && (a == 1)) && (b == 1)): %ld\n", cnt_7);
		fprintf(stderr, "OBSERVED (((c == 2) && (a == 0)) && (b == 1)): %ld\n", cnt_8);
		fprintf(stderr, "OBSERVED (((c == 2) && (a == 1)) && (b == 1)): %ld\n", cnt_9);
	}

    /* Release DMAPP resources. This is a mandatory call. */
    status = dmapp_finalize();
    if (status != DMAPP_RC_SUCCESS) {
        dmapp_explain_error(status, (const char **)&expl);
        fprintf(stderr, " dmapp_finalize FAILED w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
        exit(1);
    }

    return(0);
}

