// VPI_verbs test code
// inspired by https://thegeekinthecorner.wordpress.com/2010/09/28/rdma-read-and-write-with-ib-verbs/
#include "rdma-test.h"

extern enum mode proc;
extern struct connection self_conn;
// shared var decls
long cnt_0 = 0;
long cnt_1 = 0;
long cnt_2 = 0;
long cnt_3 = 0;
long cnt_4 = 0;
long unexpected_cnt = 0;
long long * all_failed;
long long * client_state, * server_state;
long long * Z;
long long * b_receiver;
long long * c_global;
long long * a_sender;
long long * b_sender;
long long * b_global;
long long * Y;
long long * c_sender;
long long * a_receiver;
long long * X;
long long * a_global;
long long * c_receiver;

// local decls
long long a, b, c;

void register_shared_vars(enum app_type acting_as, struct connection * conn) {
  // should be no race because the client should complete sending a request
  //   ... strictly before the server responds to it.
  if (proc == ONE_PROC && acting_as == SERVER) {
    conn->rdma_region = self_conn.rdma_region;
  } else {
    posix_memalign((void**)&(conn->rdma_region), 64, sizeof(long long)*15);
    memset(conn->rdma_region, 0, sizeof(long long)*15);
    self_conn.rdma_region = conn->rdma_region;
  }

  vars = conn->rdma_region;
  // initialize vars
  all_failed = &vars[0];
  if (app == CLIENT) {
    client_state = &vars[1];
    *client_state = 0;
    server_state = &vars[2];
  } else {
    client_state = &vars[1];
    server_state = &vars[2];
    *server_state = 0;
  }
  X = &vars[3];
  Y = &vars[4];
  Z = &vars[8];
  a_global = &vars[7];
  a_receiver = &vars[6];
  a_sender = &vars[5];
  b_global = &vars[11];
  b_receiver = &vars[10];
  b_sender = &vars[9];
  c_global = &vars[14];
  c_receiver = &vars[13];
  c_sender = &vars[12];

  test_init(conn);

  TEST_Z(conn->rdma_mr = ibv_reg_mr(pd,
				    conn->rdma_region,
				    sizeof(long long) * 15,
				    IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ | IBV_ACCESS_REMOTE_ATOMIC));
  if (acting_as == CLIENT) {
    self_conn.peer_mr = conn->rdma_mr;
  }
}

/* ****************************** START OF TESTS SECTION ****************************** */

/*
process 0 (client)
shared X = 2, Y = 0;
local a;
load(atm) a = X;

process 1 (server)
shared Z = 1;
local b, c;
Z = get(atm,atm)(Y,0);
load(atm) c = Z;
Z = cas(atm,atm,atm,atm)(X,0,Z,Z);
load(atm) b = Z;
// ---

output((c == 1) && (a == 2) && (b == 0));
output((c == 1) && (a == 2) && (b == 1));
output((c == 1) && (a == 2) && (b == 2));
output((c == 0) && (a == 2) && (b == 0));
output((c == 0) && (a == 2) && (b == 2));
*/

void test_init(struct connection *conn) {
  if (app == CLIENT) {
    *X = 2;
    *Y = 0;
  } else {
    *Z = 1;
  }
}

void test(struct ibv_cq *cq, struct connection *conn) {
  enum app_type peer;
  if (app == CLIENT) peer = SERVER; else peer = CLIENT;

  // pre-test barrier
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 1;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 1) /* busy-wait */;
    } else {
      *server_state = 1;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 1) /* busy-wait */;
    }
  }
  // test code
  if (app == CLIENT) {
    if (rand() % 2 == 1) { usleep(5 + (rand()%10)); }
    if (rand() % 2 == 1) { usleep(1); }
    a = *X;
  } else {
    if (rand() % 2 == 1) { usleep(1); }
    rdma_operation(app, conn, *conn->peer_mr, Y - vars, Z, conn->rdma_mr, IBV_WR_RDMA_READ, 0);
    if (rand() % 2 == 1) { usleep(1); }
    c = *Z;
    if (rand() % 2 == 1) { usleep(1); }
    rdma_operation_cas(app, conn, *conn->peer_mr, X - vars, Z, conn->rdma_mr, *Z, *Z);
    if (rand() % 2 == 1) { usleep(1); }
    b = *Z;
  }
  // ---
  // post-test barrier
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 2;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 2) /* busy-wait */;
    } else {
      *server_state = 2;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 2) /* busy-wait */;
    }
  }
}

void test_result(struct ibv_cq *cq, struct connection *conn) {
  enum app_type peer;
  if (app == CLIENT) peer = SERVER; else peer = CLIENT;

  if (app == CLIENT) {
    *a_global = a;
  }
  if (app == SERVER) {
    *b_global = b;
    rdma_operation(app, conn, *conn->peer_mr, b_global - vars, b_global, conn->rdma_mr, IBV_WR_RDMA_WRITE, 0);
    *c_global = c;
    rdma_operation(app, conn, *conn->peer_mr, c_global - vars, c_global, conn->rdma_mr, IBV_WR_RDMA_WRITE, 0);
  }
  send_flush(peer, cq, conn, true);
#if DUMP_VARS
  if (app == CLIENT) {
    printf("values: X=%lld Y=%lld a_sender=%lld a_receiver=%lld a_global=%lld\n", *X, *Y, *a_sender, *a_receiver, *a_global);
  }
  if (app != CLIENT) {
    printf("values:  Z=%lld b_sender=%lld b_receiver=%lld b_global=%lld c_sender=%lld c_receiver=%lld c_global=%lld\n", *Z, *b_sender, *b_receiver, *b_global, *c_sender, *c_receiver, *c_global);
  }
#endif
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 3;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 3) /* busy-wait */;
    } else {
      *server_state = 3;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 3) /* busy-wait */;
    }
  }
}

void collate_results(struct ibv_cq *cq, struct connection *conn) {
  enum app_type peer;
  if (app == CLIENT) peer = SERVER; else peer = CLIENT;


    /* Load results from global _received variables to locals */
    if (app == CLIENT) {
        a = *a_global;
        b = *b_global;
        c = *c_global;
    }

    /* Check which output is observed */
    if (app == CLIENT) {
        if (((c == 1) && (a == 2)) && (b == 0)) cnt_0++;
        else if (((c == 1) && (a == 2)) && (b == 1)) cnt_1++;
        else if (((c == 1) && (a == 2)) && (b == 2)) cnt_2++;
        else if (((c == 0) && (a == 2)) && (b == 0)) cnt_3++;
        else if (((c == 0) && (a == 2)) && (b == 2)) cnt_4++;
        else {
          fprintf(stderr, "UNEXPECTED: a=%lld b=%lld c=%lld\n", a, b, c);
          unexpected_cnt++;
        }
    }

  // synch before disconnecting
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 4;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 4) /* busy-wait */;
    } else {
      *server_state = 4;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 4) /* busy-wait */;
    }
  }
}

void print_results() {
    if (app == CLIENT) {
        fprintf(stderr, "OBSERVED (((c == 1) && (a == 2)) && (b == 0)): %ld\n", cnt_0);
        fprintf(stderr, "OBSERVED (((c == 1) && (a == 2)) && (b == 1)): %ld\n", cnt_1);
        fprintf(stderr, "OBSERVED (((c == 1) && (a == 2)) && (b == 2)): %ld\n", cnt_2);
        fprintf(stderr, "OBSERVED (((c == 0) && (a == 2)) && (b == 0)): %ld\n", cnt_3);
        fprintf(stderr, "OBSERVED (((c == 0) && (a == 2)) && (b == 2)): %ld\n", cnt_4);
        fprintf(stderr, "unexpected count %ld\n", unexpected_cnt);
    }
}

/* ****************************** END OF TESTS SECTION ****************************** */
