// VPI_verbs test code
// inspired by https://thegeekinthecorner.wordpress.com/2010/09/28/rdma-read-and-write-with-ib-verbs/
#include "rdma-test.h"

extern enum mode proc;
extern struct connection self_conn;
// shared var decls
long cnt_0 = 0;
long cnt_1 = 0;
long unexpected_cnt = 0;
long long * all_failed;
long long * client_state, * server_state;
long long * b_sender;
long long * b_global;
long long * X;
long long * a_sender;
long long * b_receiver;
long long * Z;
long long * a_receiver;
long long * Y;
long long * a_global;

// local decls
long long a, b;

void register_shared_vars(enum app_type acting_as, struct connection * conn) {
  // should be no race because the client should complete sending a request
  //   ... strictly before the server responds to it.
  if (proc == ONE_PROC && acting_as == SERVER) {
    conn->rdma_region = self_conn.rdma_region;
  } else {
    posix_memalign((void**)&(conn->rdma_region), 64, sizeof(long long)*12);
    memset(conn->rdma_region, 0, sizeof(long long)*12);
    self_conn.rdma_region = conn->rdma_region;
  }

  vars = conn->rdma_region;
  // initialize vars
  all_failed = &vars[0];
  if (app == CLIENT) {
    client_state = &vars[1];
    *client_state = 0;
    server_state = &vars[2];
  } else {
    client_state = &vars[1];
    server_state = &vars[2];
    *server_state = 0;
  }
  X = &vars[3];
  Y = &vars[4];
  Z = &vars[5];
  a_global = &vars[8];
  a_receiver = &vars[7];
  a_sender = &vars[6];
  b_global = &vars[11];
  b_receiver = &vars[10];
  b_sender = &vars[9];

  test_init(conn);

  TEST_Z(conn->rdma_mr = ibv_reg_mr(pd,
				    conn->rdma_region,
				    sizeof(long long) * 12,
				    IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ | IBV_ACCESS_REMOTE_ATOMIC));
  if (acting_as == CLIENT) {
    self_conn.peer_mr = conn->rdma_mr;
  }
}

/* ****************************** START OF TESTS SECTION ****************************** */

/*
process 0 (client)
shared X = 1, Y = 2, Z = 0;
local a, b;
put(atm,atm)(Z,0,X);
flush (0);
put(atm,atm)(Y,0,Z);
load(atm) b = Z;
load(atm) a = Y;
// ---

output((a == 1) && (b == 1));
output((a == 2) && (b == 1));
*/

void test_init(struct connection *conn) {
    *X = 1;
    *Y = 2;
    *Z = 0;
}

void test(struct ibv_cq *cq, struct connection *conn) {
  enum app_type peer;
  if (app == CLIENT) peer = SERVER; else peer = CLIENT;

  // pre-test barrier
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 1;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 1) /* busy-wait */;
    } else {
      *server_state = 1;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 1) /* busy-wait */;
    }
  }
  // test code
    if (rand() % 2 == 1) { usleep(1); }
    rdma_operation(app, conn, *conn->peer_mr, Z - vars, X, conn->rdma_mr, IBV_WR_RDMA_WRITE, 0);
    if (rand() % 2 == 1) { usleep(1); }
    send_flush(peer, cq, conn, true);
    if (rand() % 2 == 1) { usleep(1); }
    rdma_operation(app, conn, *conn->peer_mr, Y - vars, Z, conn->rdma_mr, IBV_WR_RDMA_WRITE, 0);
    if (rand() % 2 == 1) { usleep(1); }
    b = *Z;
    if (rand() % 2 == 1) { usleep(1); }
    a = *Y;
  // ---
  // post-test barrier
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 2;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 2) /* busy-wait */;
    } else {
      *server_state = 2;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 2) /* busy-wait */;
    }
  }
}

void test_result(struct ibv_cq *cq, struct connection *conn) {
  enum app_type peer;
  if (app == CLIENT) peer = SERVER; else peer = CLIENT;

  if (1) {
    *a_global = a;
    *b_global = b;
  }
  send_flush(peer, cq, conn, true);
#if DUMP_VARS
  if (app == CLIENT) {
    printf("values: X=%lld Y=%lld Z=%lld a_sender=%lld a_receiver=%lld a_global=%lld b_sender=%lld b_receiver=%lld b_global=%lld\n", *X, *Y, *Z, *a_sender, *a_receiver, *a_global, *b_sender, *b_receiver, *b_global);
  }
#endif
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 3;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 3) /* busy-wait */;
    } else {
      *server_state = 3;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 3) /* busy-wait */;
    }
  }
}

void collate_results(struct ibv_cq *cq, struct connection *conn) {
  enum app_type peer;
  if (app == CLIENT) peer = SERVER; else peer = CLIENT;


    /* Load results from global _received variables to locals */
    if (1) {
        a = *a_global;
        b = *b_global;
    }

    /* Check which output is observed */
    if (1) {
        if ((a == 1) && (b == 1)) cnt_0++;
        else if ((a == 2) && (b == 1)) cnt_1++;
        else {
          fprintf(stderr, "UNEXPECTED: a=%lld b=%lld\n", a, b);
          unexpected_cnt++;
        }
    }

  // synch before disconnecting
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 4;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 4) /* busy-wait */;
    } else {
      *server_state = 4;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 4) /* busy-wait */;
    }
  }
}

void print_results() {
    if (1) {
        fprintf(stderr, "OBSERVED ((a == 1) && (b == 1)): %ld\n", cnt_0);
        fprintf(stderr, "OBSERVED ((a == 2) && (b == 1)): %ld\n", cnt_1);
        fprintf(stderr, "unexpected count %ld\n", unexpected_cnt);
    }
}

/* ****************************** END OF TESTS SECTION ****************************** */
