// VPI_verbs test code
// inspired by https://thegeekinthecorner.wordpress.com/2010/09/28/rdma-read-and-write-with-ib-verbs/
#include "rdma-test.h"

extern enum mode proc;
extern struct connection self_conn;
// shared var decls
long cnt_0 = 0;
long unexpected_cnt = 0;
long long * all_failed;
long long * client_state, * server_state;
long long * c_receiver;
long long * d_global;
long long * d_sender;
long long * b_sender;
long long * a_sender;
long long * b_receiver;
long long * b_global;
long long * X;
long long * a_global;
long long * c_sender;
long long * c_global;
long long * a_receiver;
long long * d_receiver;

// local decls
long long a, b, c, d;

void register_shared_vars(enum app_type acting_as, struct connection * conn) {
  // should be no race because the client should complete sending a request
  //   ... strictly before the server responds to it.
  if (proc == ONE_PROC && acting_as == SERVER) {
    conn->rdma_region = self_conn.rdma_region;
  } else {
    posix_memalign((void**)&(conn->rdma_region), 64, sizeof(long long)*16);
    memset(conn->rdma_region, 0, sizeof(long long)*16);
    self_conn.rdma_region = conn->rdma_region;
  }

  vars = conn->rdma_region;
  // initialize vars
  all_failed = &vars[0];
  if (app == CLIENT) {
    client_state = &vars[1];
    *client_state = 0;
    server_state = &vars[2];
  } else {
    client_state = &vars[1];
    server_state = &vars[2];
    *server_state = 0;
  }
  X = &vars[3];
  a_global = &vars[6];
  a_receiver = &vars[5];
  a_sender = &vars[4];
  b_global = &vars[9];
  b_receiver = &vars[8];
  b_sender = &vars[7];
  c_global = &vars[12];
  c_receiver = &vars[11];
  c_sender = &vars[10];
  d_global = &vars[15];
  d_receiver = &vars[14];
  d_sender = &vars[13];

  test_init(conn);

  TEST_Z(conn->rdma_mr = ibv_reg_mr(pd,
				    conn->rdma_region,
				    sizeof(long long) * 16,
				    IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ | IBV_ACCESS_REMOTE_ATOMIC));
  if (acting_as == CLIENT) {
    self_conn.peer_mr = conn->rdma_mr;
  }
}

/* ****************************** START OF TESTS SECTION ****************************** */

/*
process 0 (client)
shared X = 0;
local a, b, c, d;
store(atm) X = 1;
load(atm) c = X;
store(atm) X = 2;
load(atm) d = X;
store(atm) X = 3;
load(atm) b = X;
load(atm) a = X;
// ---

output((d == 2) && (c == 1) && (a == 3) && (b == 3));
*/

void test_init(struct connection *conn) {
    *X = 0;
}

void test(struct ibv_cq *cq, struct connection *conn) {
  enum app_type peer;
  if (app == CLIENT) peer = SERVER; else peer = CLIENT;

  // pre-test barrier
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 1;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 1) /* busy-wait */;
    } else {
      *server_state = 1;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 1) /* busy-wait */;
    }
  }
  // test code
    if (rand() % 2 == 1) { usleep(1); }
    *X = 1;
    if (rand() % 2 == 1) { usleep(1); }
    c = *X;
    if (rand() % 2 == 1) { usleep(1); }
    *X = 2;
    if (rand() % 2 == 1) { usleep(1); }
    d = *X;
    if (rand() % 2 == 1) { usleep(1); }
    *X = 3;
    if (rand() % 2 == 1) { usleep(1); }
    b = *X;
    if (rand() % 2 == 1) { usleep(1); }
    a = *X;
  // ---
  // post-test barrier
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 2;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 2) /* busy-wait */;
    } else {
      *server_state = 2;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 2) /* busy-wait */;
    }
  }
}

void test_result(struct ibv_cq *cq, struct connection *conn) {
  enum app_type peer;
  if (app == CLIENT) peer = SERVER; else peer = CLIENT;

  if (1) {
    *a_global = a;
    *b_global = b;
    *c_global = c;
    *d_global = d;
  }
  send_flush(peer, cq, conn, true);
#if DUMP_VARS
  if (app == CLIENT) {
    printf("values: X=%lld a_sender=%lld a_receiver=%lld a_global=%lld b_sender=%lld b_receiver=%lld b_global=%lld c_sender=%lld c_receiver=%lld c_global=%lld d_sender=%lld d_receiver=%lld d_global=%lld\n", *X, *a_sender, *a_receiver, *a_global, *b_sender, *b_receiver, *b_global, *c_sender, *c_receiver, *c_global, *d_sender, *d_receiver, *d_global);
  }
#endif
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 3;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 3) /* busy-wait */;
    } else {
      *server_state = 3;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 3) /* busy-wait */;
    }
  }
}

void collate_results(struct ibv_cq *cq, struct connection *conn) {
  enum app_type peer;
  if (app == CLIENT) peer = SERVER; else peer = CLIENT;


    /* Load results from global _received variables to locals */
    if (1) {
        a = *a_global;
        b = *b_global;
        c = *c_global;
        d = *d_global;
    }

    /* Check which output is observed */
    if (1) {
        if ((((d == 2) && (c == 1)) && (a == 3)) && (b == 3)) cnt_0++;
        else {
          fprintf(stderr, "UNEXPECTED: a=%lld b=%lld c=%lld d=%lld\n", a, b, c, d);
          unexpected_cnt++;
        }
    }

  // synch before disconnecting
  if (proc == TWO_PROCS) {
    if (app == CLIENT) {
      *client_state = 4;
      rdma_operation(peer, conn, *conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*server_state < 4) /* busy-wait */;
    } else {
      *server_state = 4;
      rdma_operation(peer, conn, *conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
      consume_one_cqe(cq);
      while (*client_state < 4) /* busy-wait */;
    }
  }
}

void print_results() {
    if (1) {
        fprintf(stderr, "OBSERVED ((((d == 2) && (c == 1)) && (a == 3)) && (b == 3)): %ld\n", cnt_0);
        fprintf(stderr, "unexpected count %ld\n", unexpected_cnt);
    }
}

/* ****************************** END OF TESTS SECTION ****************************** */
