open util/integer


sig Register {
	value: one (Top+Int),
	host: one Thr
} 

/* Integer range restrictions */
//fact { all r: Register| (-1 < value[r]) and (value[r] < 3)}
//fact { all r: Reader | (-1 < rV[r]) and (rV[r] < 3)}
//fact { all w: Writer | (-1 < wV[w]) and (wV[w] < 3)}
//fact {#Thr = 2}



fact { all read: Read| host[reg[read]] = o[read] and 
								  value[reg[read]] = rV[read]}

/* undefined value */
one sig Top {}

one sig Robust {
	is_robust: one Int,
	no_conflicts: one Int
}

/* Threads */
sig Thr {}

/* Variables */
sig MemoryLocation {}

abstract sig Item {
	po : set Item,
	o: one Thr
}

sig nonAction extends Item{
}

abstract sig Action extends Item {
	/* program order, consistency order */
	co : set Action, 

	/* destination and origin thread of the action */	
	d : one Thr
	
}

/* local reads and writes */
abstract sig LocalAction extends Action {} {
}

/* external reads and writes */
abstract sig ExternalAction extends Action {}

/* local and external reads */
sig Reader in Action {
	rl: one MemoryLocation,
	W : one Writer,
	rV: one (Int + Top)
}

/* local and external writes */
sig Writer in Action {
	wl: one MemoryLocation,
	wV: one (Int + Top)
}

/* atomic statements */
sig Atomic in Action { }


/* external actions    */
abstract sig RemoteAction extends ExternalAction {
	eactions: set ExternalAction
}

abstract sig ExternAction extends ExternalAction {}

/* external read */
abstract sig ExternRead extends ExternAction {}

fact {all a: ExternRead |a in Reader and 
									not (a in Writer)}

sig AExternRead extends ExternRead {}
fact {all a: AExternRead | a in Atomic}


sig NExternRead extends ExternRead {}
fact {all a: NExternRead | not a in Atomic}


/* remote read*/
abstract sig RemoteRead extends RemoteAction {}

fact {all a: RemoteRead | a in Reader and 
										not (a in Writer)}

sig ARemoteRead extends RemoteRead {}
fact {all a: ARemoteRead | a in Atomic}

sig NRemoteRead extends RemoteRead {}
fact {all a: NRemoteRead | not a in Atomic}

/* external write */
abstract sig ExternWrite extends ExternalAction {
}

fact {all a: ExternWrite |a in Writer and 
									not (a in Reader)}

sig AExternWrite extends ExternWrite {}
fact {all a: AExternWrite | a in Atomic}

sig NExternWrite extends ExternWrite {}
fact {all a: NExternWrite | not a in Atomic}


/* remote write */
abstract sig RemoteWrite extends RemoteAction {}

fact {all a: RemoteWrite | a in Writer and 
										not (a in Reader)}

sig ARemoteWrite extends RemoteWrite {}
fact {all a: ARemoteWrite | a in Atomic}

sig NRemoteWrite extends RemoteWrite {}
fact {all a: NRemoteWrite | not a in Atomic}

/* remote read-write */

abstract sig RemoteReadWrite extends RemoteAction {}

fact {all a: RemoteReadWrite | a in Writer and a in Reader }

sig ARemoteReadWrite extends RemoteReadWrite {}
fact {all a: ARemoteReadWrite| a in Atomic}

sig NRemoteReadWrite extends RemoteReadWrite {}
fact {all a: NRemoteReadWrite | not a in Atomic }

fact {all a: RemoteReadWrite| wl[a] = rl[a]}

fact {all a: RemoteReadWrite| not a = W[a]}

/* =======  */
/* local actions */
/* =======  */

/*local write */
abstract sig Write extends LocalAction {}

fact {all a: Write | a in Writer and 
							not (a in Reader)}

sig AWrite extends Write {}
fact {all a: AWrite | a in Atomic}

sig NWrite extends Write {}
fact {all a: NWrite | not a in Atomic}

/* initial value */
sig InitialValue extends AWrite {}

fact {all i: InitialValue, a: Action-InitialValue | 
				(o[i] = d[i]) and 
				(a in co[i]) and
				(o[i] = o[a] implies a in po[i])
}

/* local read */
abstract sig Read extends LocalAction {
	reg: one Register
}

/* no same register for two read operation */
fact {all disj a, b: Read| 
				not (reg[a] = reg[b])}

fact {all a: Read | 
				a in Reader and 
				not (a in Writer)}

sig ARead extends Read {}
fact {all a: ARead | a in Atomic}

sig NRead extends Read {}
fact {all a: NRead | not a in Atomic}

/* flush statement */
sig Flush extends Action {}

fact {all a: Flush | 
				not (a in Reader) and 
				not (a in Writer)}

/* =========== */
/* === RULES === */
/* =========== */

fact { po = ^po }
fact { co = ^co }
pred poco[a, b: Action] { b in ^(po + co)[a]}

 /* consistency order is not cyclic */
fact {all a, b: Action | 
			not (a in co[b] and b in co[a])
}

 /* program order is not cyclic */
fact {all a, b: Item | 
			not (a in po[b] and b in po[a])
}

/* program order only between threads */
fact {all a, b: Item | 
			b in po[a] implies o[a] = o[b] 
}

/* reader and reads from have the same variable */
fact {all r : Reader | 
			rl[r] = wl[W[r]]
}

/* CO */
fact {all a: LocalAction, b: Action | 
					(b in po[a] implies b in co[a])
}

/* F1 */
fact {all l: LocalAction, f: Flush |
			(l in po[f] implies l in co[f]) 
}

/* F2 */
fact {all f: Flush, e: RemoteAction |
			f in po[e] and d[f] = d[e] implies 
				all a: eactions[e] | f in co[a]
}

/* F3 */
fact {all f: Flush, e: RemoteAction|
			e in po[f] and d[f] = d[e] implies 
				all a: eactions[e]| a in co[f]
}

/* R1 */
fact{all r: Reader, w: Writer|
			((not r = w) and w in co[W[r]] and wl[w] = rl[r] and r in Atomic and w in Atomic) implies
				w in co[r]
}

/* R2 */
fact {all r: Reader| 
			(r in Atomic and W[r] in Atomic) implies r in co[W[r]]
}

/* write sequentiality */
fact {all disj w1, w2 : (Writer & Atomic) | wl[w1] = wl[w2] implies (w1 in co[w2] or w2 in co[w1])}

/* In-order Routing - Optional */
fact {all e1, e2: RemoteAction | 
		(d[e1] = d[e2] and e2 in po[e1]) implies
			 e2 in co[e1]
}


/* conflicts */
pred conflictFree[r: Reader] {
	(r in Atomic) and 
	all w: Writer |
		rl[r] = wl[w] implies 
			((r in co[w] or w in co[r])
			or w in Atomic)
}

pred wConflictFree[w: Writer] {
	w in Atomic and 
	all w1: Writer - w | 
		(wl[w] = wl[w1] implies 
			((w in co[w1] or w1 in co[w])
			or w1 in Atomic))
}

/* C1 and C2 */
fact {all r: Reader |
			(conflictFree[r] and wConflictFree[W[r]]) implies
				rV[r] = wV[W[r]] else rV[r] = Top
}

/* Robustness */

/* First property */

pred first_property[r: Reader, w: Writer] {
	rl[r] = wl[w] and
	(poco[w, r] and (not r in co[w])) or 
			(poco[r, w] and  (not w in co[r])) 
}

pred second_property[r: Reader, w2: Writer] {
    rl[r] = wl[w2] and
	((poco[W[r], w2] and not (w2 in co[W[r]]))
    or 
    (poco[w2, W[r]] and not (W[r] in co[w2])))
}

fact {(all r: Reader, w: Writer | 
			(not first_property[r, w]) and (not second_property[r, w]))  implies Robust.is_robust = 1 else
				Robust.is_robust = 0
}

/* C1 and C2 */
fact {(all r: Reader |
			conflictFree[r] and wConflictFree[W[r]]) implies
				Robust.no_conflicts = 1 else Robust.no_conflicts = 0
}


pred show {#Action = 10 and #nonAction > 0 and #po[nonAction] > 0 and #RemoteReadWrite > 0 }//#Flush = 1 and #RemotePut = 1 and  #Thread = 3}

run show for 12













































