This artifact supports the OOPSLA submission "Modelling and Analysis
of Remote Memory Access Programming". Full reproducibility of this
work is challenging, as it requires access to hardware with RMA
networks. We have attempted to include all relevant code and data in
this artifact submission so that, to the extent possible, the
reviewers can verify that the artifact supports the claims in the
submission.

In particular, we include:

 - code to generate abstract test descriptions (.IR files), from our model;
 - code to generate concrete tests for DMAPP, Portals 4, and VPI Verbs, from the .IR files;
 - all of the tests that we generated; and,
 - the outputs from running the tests on our hardware.

Getting Started
===============

The easiest way to kick the tires is to use the provided Vagrant VM
image. Vagrant uses a publicly-available VM image (Debian jessie, in
our case) and provisions the image by following instructions that we
provide. The result is a VM that contains all of the parts of our
system, ready for experimentation. We include executable instructions
to create a Vagrant VM which installs Portals 4 and can run our
experiments locally against that API.

We tested our VM definition against Vagrant 1.8.1. To reproduce our Vagrant VM, do the following:

0. clearly, you've already unpacked the artifact archive `rma-alloy.tar.bz2`, since you are reading this file;
1. copy the artifact archive `rma-alloy.tar.bz2` into the directory `rma-alloy/vagrant/data/` to make it available to the Vagrant VM
2. in the `rma-alloy/vagrant/` subdirectory, run `vagrant up`
3. connect to the VM with `vagrant ssh`

You should be able to follow all of the steps below to re-generate
abstract and concrete tests, but you don't need to. You won't be able
to run the DMAPP or VPI Verbs benchmarks in this environment. 

*Important note*: the VM compiles Portals 4 for shared memory
transport, not IntelliBand transport.  You would have to install OFED
in the VM and modify the `configure` line for Portals to add
`--enable-transport-ib` to use IntelliBand. But IntelliBand is
probably not virtualized by your VM provider in any case. So you can
just see that the cases run. Don't expect any interesting behaviour.

How to run the cases
--------------------
Inside the VM, run the following commands at the prompt you get from `vagrant ssh`:

1. `export LD_LIBRARY_PATH=/usr/local/lib`
2. `tar xjvf hacking/rma-alloy/results/portals-tests.tar.bz2`
3. `cd output`
4. `./compile-and-run-portals`

(In principle, Step 1 shouldn't be required, but the ld.so cache is sometimes wrong without it.
Do it to be sure.)

The `compile-and-run-portals` script first loops through all of the
concrete cases and compiles them with `mpicc`, and then loops again
and runs them with `mpirun`. Each concrete test case is a file
`portals_test_generated.c` in its own directory. The
`compile-and-run-portals` script puts `out` files in each test case's
directory alongside the `portals_test_generated.c` file.

That concludes the quick start for the artifact. We next describe the artifact structure
and then provide instructions on how to re-generate everything.

Important files
===============

Our artifact submission includes the following files and directories:
  * directory `alloy`: the Alloy model at the heart of our submission; requires tweaks to execute directly in Alloy, but will run fine from our test generation tools
  * directory `testgen`: abstract test generation (.als to .ir)
  * directory `parser`: concrete test generation (.ir to .c)
  * file `doc/testgen.csv`: inventory of 7410 generated test cases
  * file `results/paper-irs.tar.bz2`: generated .ir files used in the paper
  * files `results/{dmapp,portals,rma}-tests.tar.bz2`: generated DMAPP, Portals 4, and VPI Verbs code
  * files `results/{dmapp,portals,rma}-tests.tar.bz2`: outputs from concrete test cases
  * directory `vagrant`: files that Vagrant can use to generate a VM with our system and Portals 4

Test generation workflow
========================

We continue by describing each of the steps in our methodology. To generate test cases and
run them on hardware, we do the following:

1. Use our `testgen` tool to create tests in Intermediate Representation (IR) using the Alloy model.
2. Use our `ConvertTo*` tools to create C code from the IR tests. Then compile the C.
3. Run the generated executables on suitable hardware and collect the results.
4. Process the results.

We have added a bonus tool, not described in the paper: an assertion
checker for litmus test programs.

Dependencies:
-------------

Should be installed system-wide:
* Java 1.8
* gradle
* antlr 4.5

In a peer directory to this software:
* alloy (http://alloy.mit.edu/alloy/downloads/alloy4.2.jar (note that it does not unpack into a subdirectory, or in the `testgen/lib` and `parser/lib` subdirectories)

To use Alloy (and we use it programmatically for test generation),
your system should be one of x86-linux, amd64-linux, x86-windows, or
x86-mac.

Step 1: testgen
---------------

In the `testgen` subdirectory, you will find a tool to generate IR testcases from the Alloy model. It is a Java application with dependencies on one of Alloy's SAT solvers.

Model driver files are found in `alloy/driver_*.als`, where * represents the rule that is being falsified for the tests to be generated.

To recreate the tests, follow these steps:

0. build with `gradle jar` in the testgen directory.
1. unpack alloy in a sibling directory to rma-alloy
2. add alloy's libraries to your 'LD_LIBRARY_PATH'; in bash, use `export LD_LIBRARY_PATH=`pwd`/../../alloy/amd64-linux/` or other suitable directory. This assumes that alloy is in indeed a sibling directory to rma-alloy and that you are on amd64-linux. Alternatively, install and use one of the other libraries, including possibly Sat4J (modify the assignment to 'options.solver' in 'ch.ethz.srl.Main').
3. in the testgen directory, run `../bin/generate-tests`.

The `bin/generate-tests` script generates all of the IR testcases from the Alloy as presented in the paper. You must run `bin/generate-tests` from the `testgen` directory. On our system, it took about 4 hours. Or, if you want to generate fewer tests, run `bin/generate-tests-small` (40 minutes). These scripts generate abstract tests in the `output` directory in our intermediate representation (.ir files). We have included the `output` directory produced by `bin/generate-tests` in `results/paper-irs.tar.bz2`.

The file `doc/testgen.csv` summarizes the behaviour of the testgen tool on our benchmarks, providing the parameters to testgen (proc and scope), the running times, and the raw and deduplicated test counts. Note that there are very minor differences in the test counts versus those reported in the paper, due to some late tweaks we made to our Alloy models.

Here is what the `generate-tests` script does. Assume that we are testing the IR rule. In testgen, run the following command:

    `java -cp build/libs/testgen.jar:lib/alloy4.2.jar ch.ethz.srl.Main --scope 10 -p 2 ../alloy/driver_IR.als`

Note that our Alloy models can't be directly run in Alloy without explicitly setting the number of threads; this is done programmatically by the testgen according to its command-line parameters.

Then, remove duplicates: 
    `fdupes -dN output`

Finally, regenerate the expected output for each test case using the IR parser and Alloy; requires `stability_check.als`:
    `cp ../parser/benchmarks/stability_check.als output`
    `java -cp ../parser/build/classes/main/:../parser/build/libs/antlr4-runtime-4.5.jar:../parser/build/libs/alloy4.2.jar Main output/driver_IR000001.ir`
    `find output -name '*.als' -delete`

NB. The testgen tool accepts the following options: --count, --first-solution, --limit, -p (number of processes) and -d (for specifying the output directory).

Step 2: ConvertTo*
------------------

In the `parser` subdirectory, you will find tools to generate C code from IR code. We have tools that generate DMAPP, Portals, and VPI Verbs (for legacy reasons, "rma") code. All of these tools require antlr 4.5. Note that the tools all expect slightly different command-line argument formats. Research code.

To recreate the tests, do the following:

0. build with 'gradle jar; gradle copyDepJars' in the parser subdirectory.

1. run with one of the following commands (DMAPP, Portals and VPI Verbs respectively):

```java -cp build/libs/parser.jar:build/libs/antlr4-runtime-4.5.jar ConvertToDmapp ../testgen/output/F1 output/paper-F1-2```

Input: all files in directory `../testgen/output/paper-F1-2` (avoid having anything but `.ir` files there!); output: in directory `output/paper-F1-2`, a `.sbatch` and a `.c` file for each input file.

Script: We have created `generate-dmapp` and `generate-dmapp-small` scripts to generate tarballs containing the `.c` and `.sbatch` files for all of the tests, assuming that you've previously run `generate-tests` or `generate-tests-small`. For your convenience, we've put the resulting tarballs in the `results` directories. You need to copy the tarball to a machine that can run them. These scripts expect to be executed from the `parser` directory and produce a tarball (`dmapp-tests.tar.bz2` or `dmapp-tests-small.tar.bz2`) in `parser/output`. Copy that tarball onto the target DMAPP machine, unpack it, and run the included `output/compile-and-run-dmapp` script to compile and run all the resulting files.

```java -cp build/libs/parser.jar:build/libs/antlr4-runtime-4.5.jar ConvertToPortals ../testgen/output/paper-F3-1/driver_F3000577.ir test```

Input: single file `../testgen/output/paper-F3-1/driver_F3000577.ir`; output: file `portals_test_generated.c` in directory `output/test`, which can be compiled with `mpicc -lportals portals_test_generated.c` (see instructions at the top of that file)

Script: While `ConvertToPortals` works differently from `ConvertToDmapp`, scripts `generate-portals` and `generate-portals-small` work the same as their DMAPP counterparts. So you can just follow the same instructions as above. The generated tarball contains script `output/compile-and-run-portals`.

`java -cp build/libs/parser.jar:build/libs/antlr4-runtime-4.5.jar ConvertToRma ../testgen/output/paper-F3-1/driver_F3000577.ir test`

Input: single file `../testgen/output/paper-F3-1/driver_F3000577.ir`; output: VPI Verbs test case in directory `output/test`, which can be compiled with the generated `Makefile`. If single-processor, invoke with `rdma-single`; if multi-processor, use `rdma-starter`.

Script: We have created `generate-rma` and `generate-rma-small` scripts to generate tarballs containing VPI Verbs code for all of the tests. Once again, you need to copy the tarball to a machine that can run them. The scripts expect to be executed from the `parser` directory and produce a tarball (`rma-tests.tar.bz2` or `rma-tests-small.tar.bz2` in its `output` subdirectory. Copy that tarball onto the target machine, and run the `output/compile-and-run-rma` script to compile and run all the resulting files.

Step 3: Running the tests
-------------------------

The generated tests tarballs contain scripts which automatically carry out instructions. 
In this section, we provide more details on how these scripts work.

Since we don't expect reviewers to have access to RMA systems, we've also included the
results from the paper in our artifact submission, under `results/*-results.tar.bz2`. Step 4
describes how we extract results from these outputs.

*DMAPP* Each generated C file is generated with a corresponding sbatch script used to submit the job to our machine:

```
#!/bin/bash -l
#
#SBATCH --time=00:30:00
#SBATCH --nodes=2
#SBATCH --ntasks-per-node=1
#SBATCH --ntasks=2
#SBATCH --output=<test_name>.out
#SBATCH --error=<test_name>.err

#=====START=====
module load slurm
echo "-------------------------------------------"
echo "The current job ID is $SLURM_JOB_ID"
echo "Running on $SLURM_JOB_NUM_NODES nodes: $SLURM_JOB_NODELIST" 
echo "Using $SLURM_NTASKS_PER_NODE tasks per node"
echo "A total of $SLURM_NTASKS tasks is used" 
echo "-------------------------------------------"
srun -n 2 -N 1 ./<test_name>
#=====END=====
```

To run a DMAPP test case:

0. copy the C file on the target machine (for us: `dora.cscs.ch`).
1. compile the C file using the 'cc' compiler
2. submit the batch script using `sbatch <script_name>`
3. monitor the status of the job using `squeue`; the interesting result is in the `<test_name>.err` file.

*Portals* This system is quite simple. Our included Vagrant VM requires `mpicc` to be run with `-DUSE_MPI`, but that is not required on the Portals system we experimented with. We use `mpirun -np 2` to execute the binary created by `mpicc` and collect the output in file `out`.

*VPI Verbs* The `compile-and-run-rma` script also builds and runs the benchmarks. Almost as simple as Portals, but we call `make` for each benchmark, and run `rdma-single` for single-process benchmarks and `rdma-starter` for two-process benchmarks.

Notes for VPI Verbs: Machine identity is hardcoded into `ConvertToRma`'s templates
for the two-machine benchmarks; it assumes that the remote machine is
called `galilei` and is reachable by passwordless ssh from the local
machine. We didn't find any additional results by actually running it
over the network. Add `-DUSE_SSH_FOR_RDMA_STARTER` to the make
invocation to enable the ssh mode; otherwise it will start the server
and client on the same machine.

Step 4: Process the results
---------------------------

For DMAPP, the useful results are in the `*.err` files scatted in
`output`'s subdirectories. For Portals and VPI Verbs, results are in
the `out` files. A couple of `grep` commands will reveal unexpected
and expected-but-not-encountered cases:

For DMAPP:

```
find . -name '*.err' | xargs grep 'UNEX'
find . -name '*.err' | xargs grep ': 0'
```

For Portals:

```
find . -name 'out' | xargs grep 'FAILED'
find . -name 'out' | xargs grep ': 0'
```

For VPI Verbs:
```
find . -name 'out' | xargs grep 'UNEX'
find . -name 'out' | xargs grep ': 0'
```

It is also possible to investigate cases that occur rarely, e.g. to find all outputs that occurred between 1 and 9 times:

```
find . -name '*.err' | xargs grep ': [1-9]$'
```

Bonus: Assertion Checker
========================

Assertion checker for litmus test programs.
To run the assetion checker, run command (from the folder parser/):

```
../bin/checker <TEST_FILE>
```

Example of a litmus test containing an assertion:

```
process 0
shared X = 1;
local a;
store(atm) X = 2;
load(atm) a = X;

process 1
shared Y = 0;
local b, c;
Y = get(atm,atm)(X,0);
Y = get(atm,atm)(X,0);
flush (0);
load(atm) b = Y;
load(atm) c = Y;
assert (!((c == 1) && (b == 2)));
```

For the example above, the checker returns:

```
Program satisfies the assertion.
```

If the assertion is not satisfied, the checker generates a counter-example and displays it using Alloy.

The assertions we handle can contain only equality operators and local variables. 
