/* main model */
open util/integer

sig Time {
	cofr, cofr_prime: Reader -> set Writer
}

one sig Order{
	First: one Time,
	Next: Time->Time
}

/* thanks to util/ordering Alloy module: models/util/ordering.als */
fun mfirst: one Time {Order.First}
fun mlast: one Time {Time - mnext.Time}
fun mnext:Time->Time {Order.Next}
fun mnexts[t: Time]: set Time {t.^(Order.Next)}
pred mlt [t1, t2: Time] {t2 in mnexts[t1]}

fact {all t: Time| not mlt[t, Order.First]}
fact {all disj t1, t2: Time| mlt[t1, t2] or mlt[t2, t1]}
fact {no t: Time| mlt[t, t]}

/* initially, cofr is empty */
fact {
	all r: Reader| #cofr[mfirst, r] = 0 and #cofr_prime[mfirst, r] = 0
}

fact {
	no disj t, t': Time| t' in mnexts[t] and cofr[t']  in cofr[t] 
}

/* undefined value */
one sig Top {}

abstract sig Boolean {}
one sig True, False extends Boolean {}

one sig Robust {
	no_conflicts: one Boolean
}

/* Locals */
sig Register {
	host: one Thr
} 

fact { all r: Register| host[r] = o[~reg[r]] }

/* Threads */
sig Thr {}

/* Variables */
sig MemoryLocation {
  host: one Thr
}

abstract sig Item {}


abstract sig Statement {
	actions: set ExternalAction
}

/* =============== */
/* Remote Put statement */
/* =============== */

sig Rp extends Statement {}

fact { all r: Rp | #r.actions = 2 and #(r.actions & ExternRead) = 1 and #(r.actions & RemoteWrite) = 1}

/* =============== */
/* Remote Get statement */
/* =============== */

sig Rg extends Statement {}

fact { all r: Rg | #r.actions = 2 and 
							#(r.actions & RemoteRead) = 1 and 
							#(r.actions & ExternWrite) = 1}


/* ======================= */
/* Remote Get Accumulate statement */
/* ======================= */

sig Rga extends Statement {}

fact { all r: Rga | #r.actions = 3 and 
							#(r.actions & ExternWrite) = 1 and 
							#(r.actions & ExternRead) = 1 and 
							#(r.actions & RemoteReadWrite) = 1
}

/* ========================= */
/* Remote Compare and Swap statement */
/* ========================= */

sig Rcas extends Statement {}

fact {all r: Rcas | #r.actions = 4 and 
							#(r.actions & ExternRead) = 2 and 
							#(r.actions & ExternWrite) = 1 and
							#(r.actions & RemoteReadWrite) = 1}

abstract sig Action extends Item {
	/* program order, consistency order */
	po : set Action, 

	/* destination and origin thread of the action */	
	d, o : one Thr
	
}

/* local reads and writes */
abstract sig LocalAction extends Action {
	poico: set Action
} 

fact { all l: LocalAction| o[l] = d[l]}

/* external reads and writes */
abstract sig ExternalAction extends Action {
	coexternalflush: set Flush
}


/* local and external reads */
sig Reader in Action {
	rl: one MemoryLocation,
	W : one Writer,
	rV: one (Int + Top)
}

fact {all r: Reader| d[r] = host[rl[r]] }

/* local and external writes */
sig Writer in Action {
	wl: one MemoryLocation,
	wV: one (Int + Top),

	corf: set Reader,
	cows: set Writer
}

fact {all w: Writer| d[w] = host[wl[w]]}

/* ============ */
/* atomic statements */
/* ============ */
sig Atomic in Action { }

/* =========== */
/* external actions  */
/* =========== */
abstract sig RemoteAction extends ExternalAction {
	eactions: set ExternalAction,
	coir: set RemoteAction
}

abstract sig ExternAction extends ExternalAction {}

fact {all e : ExternAction | d[e] = o[e]}


/* ========= */
/* extern read */
/* ========= */
abstract sig ExternRead extends ExternAction {

	coput: lone RemoteWrite,
	corga1: lone RemoteReadWrite,
	cocas1: lone RemoteReadWrite	
}

fact {all a: ExternRead |
				a in Reader and 
				not (a in Writer)}

sig AExternRead extends ExternRead {}
fact {all a: AExternRead | a in Atomic}


sig NExternRead extends ExternRead {}
fact {all a: NExternRead | not a in Atomic}

/* ======== */
/* remote read */
/* ======== */
abstract sig RemoteRead extends RemoteAction {
	coget: one ExternWrite
}

fact {all a: RemoteRead |
				a in Reader and 
				not (a in Writer)}

sig ARemoteRead extends RemoteRead {}
fact {all a: ARemoteRead | a in Atomic}

sig NRemoteRead extends RemoteRead {}
fact {all a: NRemoteRead | not a in Atomic}

/* ========= */
/* extern write */
/* ========= */
abstract sig ExternWrite extends ExternAction {
}

fact {all a: ExternWrite |
				a in Writer and 
				not (a in Reader)}

sig AExternWrite extends ExternWrite {}
fact {all a: AExternWrite | a in Atomic}

sig NExternWrite extends ExternWrite {}
fact {all a: NExternWrite | not a in Atomic}

/* ======== */
/* remote write */
/* ======== */
abstract sig RemoteWrite extends RemoteAction {}

fact {all a: RemoteWrite |
				a in Writer and 
				not (a in Reader)}

sig ARemoteWrite extends RemoteWrite {}
fact {all a: ARemoteWrite | a in Atomic}

sig NRemoteWrite extends RemoteWrite {}
fact {all a: NRemoteWrite | not a in Atomic}

/* ============= */
/* remote read - write */
/* ============= */

abstract sig RemoteReadWrite extends RemoteAction{
	corga2: lone ExternWrite,
	cocas2: lone ExternWrite
}

fact {all a : RemoteReadWrite | 
			a in Writer and 
			a in Reader}

sig ARemoteReadWrite extends RemoteReadWrite {}
fact {all a: ARemoteReadWrite | a in Atomic}

sig NRemoteReadWrite extends RemoteReadWrite {}
fact {all a: NRemoteReadWrite | not a in Atomic}

fact {all a: RemoteReadWrite | rl[a] = wl[a]}

/* =======  */
/* local actions */
/* =======  */

/* ====== */
/*local write */
/* ====== */
abstract sig Write extends LocalAction {}

fact {all a: Write | 
				a in Writer and 
				not (a in Reader)}

sig AWrite extends Write {}
fact {all a: AWrite | a in Atomic}

sig NWrite extends Write {}
fact {all a: NWrite | not a in Atomic}

/* ======= */
/* initial value */
/* ======= */ 
sig InitialValue extends AWrite {
	coinit: set Action-InitialValue
}

fact {all i: InitialValue, a: Action-InitialValue | 
				(a in coinit[i]) and
				(o[i] = o[a] implies a in po[i])
}

/* ====== */ 
/* local read */
/* ====== */
abstract sig Read extends LocalAction {
	reg: one Register
}

fact {all disj a, b: Read|
				not (reg[a] = reg[b])}

fact {all a: Read | 
				a in Reader and 
				not (a in Writer)}

sig ARead extends Read {}
fact {all a: ARead | a in Atomic}

sig NRead extends Read {}
fact {all a: NRead | not a in Atomic}

/* ========== */
/* flush statement */
/* ========== */
sig Flush extends Action {
	coflushlocal: set LocalAction,
	coflushexternal: set ExternalAction	
}

fact {all a: Flush | 
				not (a in Reader) and 
				not (a in Writer) and
				not (a in Atomic)}

