open actions as a

/* ================= */
/* Remote Statements rules */
/* ================= */

/* There are no actions in between (PO) actions generated by remote statements */
fact {all disj e1, e2: ExternalAction, a: Action | 
		(~actions[e1] = ~actions[e2]) implies ((~actions[a] = ~actions[e1]) or (not (e1 in po[a] and a in po[e2])))
}

/* All actions corresponding to a statement have the same origin thread */
fact { all s: Statement| #(s.actions.o) = 1}

/* All ExternalAction are triggered by a Statement */
fact {Statement.actions = ExternalAction}

/* Statements cannot share actions */
fact {all disj s1, s2: Statement | #(s1.actions & s2.actions) = 0}


/* =========== */
/* Remote Put rules */
/* =========== */

fact {all disj r: ExternRead, w: RemoteWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rp) iff 
			w in coput[r]
}

fact {all disj r: ExternRead, w: RemoteWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rp) iff 
			(rV[r] = wV[w] and
				w in po[r] and
			 	r = eactions[w])
}

/* =========== */
/* Remote Get rules */
/* =========== */

fact {all disj r: RemoteRead, w: ExternWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rg) iff 
			w in coget[r]
}

fact {all disj r: RemoteRead, w: ExternWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rg) iff 
			(rV[r] = wV[w] and 
				w in po[r] and 
				w = eactions[r])
}

/* =================== */
/* Remote Get Accumulate rules */
/* =================== */

fact {all disj r: ExternRead, rw: RemoteReadWrite |
		(~actions[r] = ~actions[rw] and 
		 ~actions[r] in Rga)
		 iff (rw in corga1[r])
}

fact {all disj w: ExternWrite, rw: RemoteReadWrite |
		(~actions[w] = ~actions[rw] and 
		 ~actions[w] in Rga)
		 iff (w in corga2[rw])
}

fact {all disj r: ExternRead, w: ExternWrite, rw: RemoteReadWrite |
		(~actions[r] = ~actions[w] and 
		 ~actions[w] = ~actions[rw] and 
		 ~actions[w] in Rga)
		 iff 
			/* fixed RGA return value*/
			(wV[w] = rV[rw] and 
			 wV[rw] = rV[r].plus[rV[rw]] and 

			 (rw in po[r]) and 
			 (w in po[rw]) and 

			 (r in eactions[rw]) and 
			 (w in eactions[rw]) and 
			 #eactions[rw] = 2)
}

/* ===================== */
/* Remote Compare and Swap rules */
/* ===================== */

fact {all disj r: ExternRead, rw: RemoteReadWrite|

		(~actions[r] = ~actions[rw] and
		 ~actions[rw] in Rcas)
		iff
		rw in cocas1[r]
}

fact {all disj w: ExternWrite, rw: RemoteReadWrite|

		(~actions[rw] = ~actions[w] and
		 ~actions[rw] in Rcas)
		iff
		w in cocas2[rw]
}

fact {all disj rcomp, rnew: ExternRead, w: ExternWrite, rw: RemoteReadWrite|

		(~actions[rcomp] = ~actions[rnew] and
		 ~actions[rnew] = ~actions[rw] and
		 ~actions[rw] = ~actions[w] and
		 rnew in po[rcomp] and
		 ~actions[rw] in Rcas)
		iff
			(rcomp in eactions[rw] and
			 rnew in eactions[rw] and
			 w in eactions[rw] and
			 #eactions[rw] = 3 and

			 rw in po[rnew] and
			 w in po[rw] and 

			( (rV[rcomp] = rV[rw]) implies
				(wV[rw] = rV[rnew]) else
				(wV[rw] = rV[rw]) ) and

			wV[w] = rV[rw]) }

/* ========= */
/* Relations rules */
/* ========= */

/* PO is transitively closed */
fact { po = ^po }

/* program order only between threads */
fact {all disj a, b: Action | 
			(b in po[a] or a in po[b]) iff o[a] = o[b] 
}

/* reader and reads from have the same variable */
fact {all r : Reader | 
			rl[r] = wl[W[r]] and d[r] = d[W[r]]
}

/* CO rule (poico) */
fact {all a: LocalAction, b: Action | 
					(b in po[a] iff b in poico[a])
}

/* F1 rule (coflushlocal) */
fact {all l: LocalAction, f: Flush |
			(l in po[f] iff l in coflushlocal[f]) 
}

/* F3 rule (coflushexternal) */
fact {all f: Flush, e: RemoteAction|
//			(e in po[f] and d[f] = d[e]) iff 
			/* Make the flush total, as it is in DMAPP */
			(e in po[f]) iff 
				(eactions[e] in coflushexternal[f] and e in coflushexternal[f])
}


/* Write Sequentiality */
/*fact {all w1, w2: Writer| 
		((not (w1 = w2)) and (wl[w1] = wl[w2]) and (w1 in Atomic) and (w2 in Atomic)) iff 
		   ((w1 in cows[w2] and not w2 in cows[w1]) or (w2 in cows[w1] and not w1 in cows[w2])) }*/

/* conflicts */
pred conflictFree[r: Reader] {
	(r in Atomic) and 
	all w: Writer |
		rl[r] = wl[w] implies 
			((co[w, r, mlast] or co[r, w, mlast])
			or w in Atomic)
}

pred wConflictFree[w: Writer] {
	w in Atomic and 
	all w1: Writer - w | 
		(wl[w] = wl[w1] implies 
			((co[w1, w, mlast] or co[w, w1, mlast])
			or w1 in Atomic))
}

/* C1 and C2 rules */
fact {all r: Reader |
			(conflictFree[r] and wConflictFree[W[r]]) implies
				rV[r] = wV[W[r]] else rV[r] = Top
}

/* conflict freedom  */
fact {(all r: Reader |
			conflictFree[r] and wConflictFree[W[r]]) implies
				Robust.no_conflicts = True else Robust.no_conflicts = False
}

pred acyclic[r: Action -> Action] {
	no a: Action| a in r[a]
}

/* Program order is acyclic */
fact { acyclic[po] }

/* consistency order predicate */
pred co[a, b: Action, t: Time] {
	b in ^(poico + coflushlocal + coexternalflush+ cocas1 + cocas2 + coir + 
		coflushexternal + corf + cofr[t] + coput + //cows + 
		coget + corga1 + corga2 + coinit)[a]
}


