open test_rules_IR as c

/* Generate initial values for all variables */
fact {MemoryLocation = wl[InitialValue] }

/* Exists at most one initial value for each variable */
fact { all disj i, j: InitialValue | not wl[i] = wl[j]} 

/* Values written by local writes are increasing with CO */
fact { all disj w1, w2: Write| (co[w1, w2, mlast]) implies wV[w2] > wV[w1]}

/* Generate distinct write values for initial values and local writes */
fact { all disj w1, w2: Write| not (wV[w1] = wV[w2])}

/* Values written by local writes are between [0 .. #Write]*/
fact { all w: Write| 0 <=  wV[w] and wV[w] < #Write}

/* There should be at least one local read from each writer (except initial values)*/
fact {W[Read] = Writer - InitialValue}

/* ReaderPivot is always a local Read */
fact { ((FirstPivot in Reader) and (SecondPivot in Writer)) implies 
					(rl[FirstPivot] = wl[SecondPivot] and not rV[FirstPivot] = wV[SecondPivot])}

fact { ((FirstPivot in Writer) and (SecondPivot in Reader)) implies 
					(wl[FirstPivot] = rl[SecondPivot] and not wV[FirstPivot] = rV[SecondPivot])}

fact { ((FirstPivot in Writer) and (SecondPivot in Writer)) implies 
					(wl[FirstPivot] = wl[SecondPivot] and not wV[FirstPivot] = wV[SecondPivot])}

fact {  ((FirstPivot in Reader) and (SecondPivot in Reader)) implies 
					(rl[FirstPivot] = rl[SecondPivot] and not rV[FirstPivot] = rV[SecondPivot])}


one sig TheFlush in Flush{}

fact {TheFlush in po[SecondPivot] and d[TheFlush] = d[SecondPivot]}

fact {no l: Read| l in po[SecondPivot] and TheFlush in po[l]}


/* no two consecutive flushes (without a Remote Action in between) */
fact {no disj f1, f2: Flush| ((o[f1] = o[f2]) and (d[f1] = d[f2]) and 
											(no r: RemoteAction| r in po[f1] and f2 in po[r])) }

/* a flush should always come after a RemoteAction */
fact {no f: Flush| (no r: RemoteAction| (d[f] = d[r]) and (f in po[r]))}

/* read from all variables */
fact { rl[Reader] = MemoryLocation }

/* eliminate unused threads */
fact {Thr = o[Action]}

/* generate only atomic actions*/
fact {(Action-Flush) in Atomic}

/* testgen selects one of these... */
pred oneThread { #Thr = 1 }
pred twoThreads { #Thr = 2 }

/* each thread has at least a writer action */
fact {all t: Thr| #(~o[t] & Writer) > 0}

/* eliminate remote = origin */
fact {(#Thr = 2) implies (all r: RemoteAction| not o[r] = d[r])}

/* Properties of the test */
pred show { 
            //#(Action.o) > 1 and
            //#Rcas = 0 and
            //#Rga = 0 and
            //#Action = 7 and
			//#Rga = 1 and
            Robust.no_conflicts = True }

/* removes Univ objects */
sig Dummy {}
fact { no Dummy }

run show for 10
