/* main model */
open util/integer
open util/ordering[Time] as to

sig Time {
	cofr, cofr_prime: Reader -> set Writer
}

/* initially, cofr is empty */
fact {
	all r: Reader| #cofr[to/first, r] = 0 and #cofr_prime[to/first, r] = 0
}

/* undefined value */
one sig Top {}

abstract sig Boolean {}
one sig True, False extends Boolean {}

one sig Robust {
	is_robust: one Boolean,
	no_conflicts: one Boolean
}

/* Locals */
sig Register {
	host: one Thr
} 

fact { all r: Register| host[r] = o[~reg[r]] }

/* Threads */
sig Thr {}

/* Variables */
sig MemoryLocation {
  host: one Thr
}

abstract sig Item {}


abstract sig Statement {
	actions: set ExternalAction
}

/* =============== */
/* Remote Put statement */
/* =============== */

sig Rp extends Statement {}

fact { all r: Rp | #r.actions = 2 and #(r.actions & ExternRead) = 1 and #(r.actions & RemoteWrite) = 1}

fact {all disj r: ExternRead, w: RemoteWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rp) iff 
			w in coput[r]
}

fact {all disj r: ExternRead, w: RemoteWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rp) iff 
			(rV[r] = wV[w] and
				w in po[r] and
			 	r = eactions[w])
}

fact {all disj e1, e2: ExternalAction, a: Action | 
		(~actions[e1] = ~actions[e2]) implies ((~actions[a] = ~actions[e1]) or (not (e1 in po[a] and a in po[e2])))
}


/* =============== */
/* Remote Get statement */
/* =============== */

sig Rg extends Statement {}
fact { all r: Rg | #r.actions = 2 and 
							#(r.actions & RemoteRead) = 1 and 
							#(r.actions & ExternWrite) = 1}

fact {all disj r: RemoteRead, w: ExternWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rg) iff 
			w in coget[r]
}

fact {all disj r: RemoteRead, w: ExternWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rg) iff 
			(rV[r] = wV[w] and 
				w in po[r] and 
				w = eactions[r])
}


/* ======================= */
/* Remote Get Accumulate statement */
/* ======================= */

sig Rga extends Statement {}

fact { all r: Rga | #r.actions = 3 and 
							#(r.actions & ExternWrite) = 1 and 
							#(r.actions & ExternRead) = 1 and 
							#(r.actions & RemoteReadWrite) = 1
}

fact {all disj r: ExternRead, rw: RemoteReadWrite |
		(~actions[r] = ~actions[rw] and 
		 ~actions[r] in Rga)
		 iff (rw in corga1[r])
}

fact {all disj w: ExternWrite, rw: RemoteReadWrite |
		(~actions[w] = ~actions[rw] and 
		 ~actions[w] in Rga)
		 iff (w in corga2[rw])
}

fact {all disj r: ExternRead, w: ExternWrite, rw: RemoteReadWrite |
		(~actions[r] = ~actions[w] and 
		 ~actions[w] = ~actions[rw] and 
		 ~actions[w] in Rga)
		 iff 
			(wV[rw] = wV[w] and 
			 wV[rw] = rV[r].plus[rV[rw]] and 

			 (rw in po[r]) and 
			 (w in po[rw]) and 

			 (r in eactions[rw]) and 
			 (w in eactions[rw]) and 
			 #eactions[rw] = 2)
}

//fact {all disj r1, r2: Reader, w: Writer, rw: (Reader & Writer) | 1 = 1}

/* ========================= */
/* Remote Compare and Swap statement */
/* ========================= */

sig Rcas extends Statement {}

fact {all r: Rcas | #r.actions = 4 and 
							#(r.actions & ExternRead) = 2 and 
							#(r.actions & ExternWrite) = 1 and
							#(r.actions & RemoteReadWrite) = 1}


fact {all disj r: ExternRead, rw: RemoteReadWrite|

		(~actions[r] = ~actions[rw] and
		 ~actions[rw] in Rcas)
		iff
		rw in cocas1[r]
}

fact {all disj w: ExternWrite, rw: RemoteReadWrite|

		(~actions[rw] = ~actions[w] and
		 ~actions[rw] in Rcas)
		iff
		w in cocas2[rw]
}

fact {all disj rcomp, rnew: ExternRead, w: ExternWrite, rw: RemoteReadWrite|

		(~actions[rcomp] = ~actions[rnew] and
		 ~actions[rnew] = ~actions[rw] and
		 ~actions[rw] = ~actions[w] and
		 rnew in po[rcomp] and
		 ~actions[rw] in Rcas)
		iff
			(rcomp in eactions[rw] and
			 rnew in eactions[rw] and
			 w in eactions[rw] and
			 #eactions[rw] = 3 and

			 rw in po[rnew] and
			 w in po[rw] and 

			( (rV[rcomp] = rV[rw]) implies
				(wV[rw] = rV[rnew]) else
				(wV[rw] = rV[rw]) ) and

			wV[w] = rV[rw]) }



/* All actions corresponding to a statement have the same origin thread */
fact { all s: Statement| #(s.actions.o) = 1}

/* All ExternalAction are triggered by a Statement */
fact {Statement.actions = ExternalAction}

/* Statements cannot share actions */
fact {all disj s1, s2: Statement | #(s1.actions & s2.actions) = 0}



abstract sig Action extends Item {
	/* program order, consistency order */
	po : set Action, 

	/* destination and origin thread of the action */	
	d, o : one Thr
	
}

/* local reads and writes */
abstract sig LocalAction extends Action {
	poico: set Action
} 

fact { all l: LocalAction| o[l] = d[l]}

/* external reads and writes */
abstract sig ExternalAction extends Action {
	coexternalflush: set Flush
}


/* local and external reads */
sig Reader in Action {
	rl: one MemoryLocation,
	W : one Writer,
	rV: one (Int + Top)
}

fact {all r: Reader| d[r] = host[rl[r]] }

/* local and external writes */
sig Writer in Action {
	wl: one MemoryLocation,
	wV: one (Int + Top),

	corf: set Reader,
	cows: set Writer
}

fact {all w: Writer| d[w] = host[wl[w]]}

/* ============ */
/* atomic statements */
/* ============ */
sig Atomic in Action { }

/* =========== */
/* external actions  */
/* =========== */
abstract sig RemoteAction extends ExternalAction {
	eactions: set ExternalAction
}

abstract sig ExternAction extends ExternalAction {}

fact {all e : ExternAction | d[e] = o[e]}


/* ========= */
/* extern read */
/* ========= */
abstract sig ExternRead extends ExternAction {

	coput: lone RemoteWrite,
	corga1: lone RemoteReadWrite,
	cocas1: lone RemoteReadWrite	
}

fact {all a: ExternRead |
				a in Reader and 
				not (a in Writer)}

sig AExternRead extends ExternRead {}
fact {all a: AExternRead | a in Atomic}


sig NExternRead extends ExternRead {}
fact {all a: NExternRead | not a in Atomic}

/* ======== */
/* remote read */
/* ======== */
abstract sig RemoteRead extends RemoteAction {
	coget: one ExternWrite
}

fact {all a: RemoteRead |
				a in Reader and 
				not (a in Writer)}

sig ARemoteRead extends RemoteRead {}
fact {all a: ARemoteRead | a in Atomic}

sig NRemoteRead extends RemoteRead {}
fact {all a: NRemoteRead | not a in Atomic}

/* ========= */
/* extern write */
/* ========= */
abstract sig ExternWrite extends ExternAction {
}

fact {all a: ExternWrite |
				a in Writer and 
				not (a in Reader)}

sig AExternWrite extends ExternWrite {}
fact {all a: AExternWrite | a in Atomic}

sig NExternWrite extends ExternWrite {}
fact {all a: NExternWrite | not a in Atomic}

/* ======== */
/* remote write */
/* ======== */
abstract sig RemoteWrite extends RemoteAction {}

fact {all a: RemoteWrite |
				a in Writer and 
				not (a in Reader)}

sig ARemoteWrite extends RemoteWrite {}
fact {all a: ARemoteWrite | a in Atomic}

sig NRemoteWrite extends RemoteWrite {}
fact {all a: NRemoteWrite | not a in Atomic}

/* ============= */
/* remote read - write */
/* ============= */

abstract sig RemoteReadWrite extends RemoteAction{
	corga2: lone ExternWrite,
	cocas2: lone ExternWrite
}

fact {all a : RemoteReadWrite | 
			a in Writer and 
			a in Reader}

sig ARemoteReadWrite extends RemoteReadWrite {}
fact {all a: ARemoteReadWrite | a in Atomic}

sig NRemoteReadWrite extends RemoteReadWrite {}
fact {all a: NRemoteReadWrite | not a in Atomic}

fact {all a: RemoteReadWrite | rl[a] = wl[a]}

/* =======  */
/* local actions */
/* =======  */

/* ====== */
/*local write */
/* ====== */
abstract sig Write extends LocalAction {}

fact {all a: Write | 
				a in Writer and 
				not (a in Reader)}

sig AWrite extends Write {}
fact {all a: AWrite | a in Atomic}

sig NWrite extends Write {}
fact {all a: NWrite | not a in Atomic}

/* ======= */
/* initial value */
/* ======= */
sig InitialValue extends AWrite {
	coinit: set Action-InitialValue
}

fact {all i: InitialValue, a: Action-InitialValue | 
				(a in coinit[i]) and
				(o[i] = o[a] implies a in po[i])
}

/* ====== */
/* local read */
/* ====== */
abstract sig Read extends LocalAction {
	reg: one Register
}

fact {all disj a, b: Read|
				not (reg[a] = reg[b])}

fact {all a: Read | 
				a in Reader and 
				not (a in Writer)}

sig ARead extends Read {}
fact {all a: ARead | a in Atomic}

sig NRead extends Read {}
fact {all a: NRead | not a in Atomic}

/* ========== */
/* flush statement */
/* ========== */
sig Flush extends Action {
	coflushlocal: set LocalAction,
	coflushexternal: set ExternalAction	
}

fact {all a: Flush | 
				not (a in Reader) and 
				not (a in Writer)}

/* ===== */
/*  RULES  */
/* ===== */

fact { po = ^po }

//fact { poco = ^(po + co)}



/* program order only between threads */
fact {all disj a, b: Action | 
			(b in po[a] or a in po[b]) iff o[a] = o[b] 
}

/* reader and reads from have the same variable */
fact {all r : Reader | 
			rl[r] = wl[W[r]] and d[r] = d[W[r]]
}

/* CO => poico*/
fact {all a: LocalAction, b: Action | 
					(b in po[a] iff b in poico[a])
}

/* F1 => coflushlocal */
fact {all l: LocalAction, f: Flush |
			(l in po[f] iff l in coflushlocal[f]) 
}

/* F2 => coexternflush */
fact {all f: Flush, e: RemoteAction |
			(f in po[e] and d[f] = d[e]) iff 
				(f in coexternalflush[e] and all a: eactions[e] | f in coexternalflush[a])
}

/* F3 => coflushexternal*/
fact {all f: Flush, e: RemoteAction|
			(e in po[f] and d[f] = d[e]) iff 
				(eactions[e] in coflushexternal[f] and e in coflushexternal[f])
}

/* R2 => corf */
fact {all r: Reader| 
			(r in Atomic and W[r] in Atomic) iff 
				(r in corf[W[r]])
}

/* cows */
fact {all w1, w2: Writer| 
			((not (w1 = w2)) and (wl[w1] = wl[w2]) and (w1 in Atomic) and (w2 in Atomic)) iff 
			   ((w1 in cows[w2]) or (w2 in cows[w1])) }

/* In-order Routing - Optional */
/*fact {all e1, e2: RemoteAction | 
		(d[e1] = d[e2] and e2 in po[e1]) implies
			 e2 in co[e1]
}*/


/* conflicts */
pred conflictFree[r: Reader] {
	(r in Atomic) and 
	all w: Writer |
		rl[r] = wl[w] implies 
			((co[w, r, to/last] or co[r, w, to/last])
			or w in Atomic)
}

pred wConflictFree[w: Writer] {
	w in Atomic and 
	all w1: Writer - w | 
		(wl[w] = wl[w1] implies 
			((co[w1, w, to/last] or co[w, w1, to/last])
			or w1 in Atomic))
}

/* C1 and C2 */
fact {all r: Reader |
			(conflictFree[r] and wConflictFree[W[r]]) implies
				rV[r] = wV[W[r]] else rV[r] = Top
}

/* Robustness */

/* First robustness criterion */
/*pred first_property[r: Reader, w: Writer] {
	rl[r] = wl[w] and
	(poco[w, r] and (not co[w, r])) or 
			(poco[r, w] and  (not co[r, w])) 
}*/ 

/* Second robustness criterion */
/*pred second_property[r: Reader, w2: Writer] {
    rl[r] = wl[w2] and
	((poco[W[r], w2] and not (co[W[r], w2]))
    or 
    (poco[w2, W[r]] and not (co[w2, W[r]])))
}*/

/* Robustness rule */
/*fact {(all r: Reader, w: Writer | not first_property[r, w] and not second_property[r, w])  implies 
												Robust.is_robust = True else
												Robust.is_robust = False
}*/

/* C1 and C2 */
fact {(all r: Reader |
			conflictFree[r] and wConflictFree[W[r]]) implies
				Robust.no_conflicts = True else Robust.no_conflicts = False
}

one sig ReaderPivot in Reader{} 

one sig WriterPivot in Writer{}

fact { not ReaderPivot = WriterPivot }

/* consistency order without R1 on Pivots */
pred co_prime[a, b: Action, t: Time] {
	 b in ^(poico + coflushlocal + coexternalflush+ cocas1 + cocas2 +
		coflushexternal + corf + cofr_prime[t] + coput +
		coget + corga1 + corga2 + coinit)[a]
}

/* consistency order predicate */
pred co[a, b: Action, t: Time] {
	b in ^(poico + coflushlocal + coexternalflush+ cocas1 + cocas2 +
		coflushexternal + corf + cofr[t] + coput +
		coget + corga1 + corga2 + coinit)[a]
}

pred R1_hypothesis_co[r: Action, w: Action, t: Time] {
			(not (r = w)) and
			(r in (Reader & Atomic)) and
			(w in (Writer & Atomic)) and
			(wl[w] = rl[r]) and 
			co[W[r], w, t] and
			(not (w = W[r]))
}

pred R1_hypothesis_co_prime[r: Action, w: Action, t: Time] {
			(not (r = w)) and
			(r in (Reader & Atomic)) and
			(w in (Writer & Atomic)) and
			(wl[w] = rl[r]) and 
			co_prime[W[r], w, t] and
			(not (w = W[r]))
}

/* R1 => cofr */
fact{all disj r, w: Action, disj t, t': Time |
			lt[t, t'] implies ((R1_hypothesis_co[r, w, t]) iff 
				w in cofr[t', r]) 
}

/* (R1 - Pivot) => cofr_prime */
fact{all disj r, w: Action, disj t, t': Time |
			(lt[t, t']  and not ((r = ReaderPivot) and (w = WriterPivot)) ) implies 
				((R1_hypothesis_co_prime[r, w, t] ) iff 
					w in cofr_prime[t', r])
}

fact{R1_hypothesis_co[ReaderPivot, WriterPivot, to/last]}

/* The writer pivot writes different value from the W[the reader pivot] */
 fact{ not wV[WriterPivot] = wV[W[ReaderPivot]]}


/* Do not generate Top write values */
fact {all w: Write | not wV[w] = Top}

pred acyclic[r: Action -> Action] {
	no a: Action| a in r[a]
}

/* Program order is acyclic */
fact { acyclic[po] }

/* CO' is acyclic */
fact { no a: Action | co_prime[a, a, to/last]}

/* CO is cyclic, and the reader pivot is involved in the cycle */
fact {co[ReaderPivot, ReaderPivot, to/last]}

/* Generate initial values for all variables */
fact {MemoryLocation = wl[InitialValue] }

/* Exists at most one initial value for each variable */
fact { all disj i, j: InitialValue | not wl[i] = wl[j]}

/* Generate distinct write values for initial values and local writes */
fact { all disj w1, w2: Write| not (wV[w1] = wV[w2])}

/* Values written by local writes are increasing with CO */
fact { all disj w1, w2: Write| (co[w1, w2, to/last]) implies wV[w2] > wV[w1]}

/* Values written by local writes are between [0 .. #Write]*/
fact { all w: Write| 0 <=  wV[w] and wV[w] < #Write}


/* Witness */
one sig Witness in Read {}

/* Witness reads the same variable as the reader Pivot*/
fact{ rl[Witness] = rl[ReaderPivot]}

/* no two consecutive flushes (without a Remote Action in between) */
fact {no disj f1, f2: Flush| ((o[f1] = o[f2]) and (d[f1] = d[f2]) and 
											(no r: RemoteAction| r in po[f1] and f2 in po[r])) }

/* a flush should always come after a RemoteAction */
fact {no f: Flush| (no r: RemoteAction| (d[f] = d[r]) and (f in po[r]))}

/* read from all variables */
fact { rl[Reader] = MemoryLocation }
