open util/integer

/* undefined value */
one sig Top {}

abstract sig Boolean {}
one sig True, False extends Boolean {}

one sig Robust {
	is_robust: one Boolean,
	no_conflicts: one Boolean
}

/* Locals */
sig Register {
	host: one Thr
} 

fact { all r: Register| host[r] = o[~reg[r]]

}

/* Threads */
sig Thr {}

/* Variables */
sig MemoryLocation {
  host: one Thr
}

abstract sig Item {}


abstract sig Statement {
	actions: set ExternalAction
}

/* =============== */
/* Remote Put statement */
/* =============== */

sig Rp extends Statement {}

fact { all r: Rp | #r.actions = 2 and #(r.actions & ExternRead) = 1 and #(r.actions & RemoteWrite) = 1}

fact {all disj r: ExternRead, w: RemoteWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rp) iff 
			w in coput[r]
}

fact {all disj r: ExternRead, w: RemoteWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rp) iff 
			(rV[r] = wV[w] and w in po[r] and
			 r = eactions[w])
}

fact {all disj e1, e2: ExternalAction, a: Action | 
		~actions[e1] = ~actions[e2] implies not (e1 in po[a] and a in po[e2])
}


/* =============== */
/* Remote Get statement */
/* =============== */

sig Rg extends Statement {}
fact { all r: Rg | #r.actions = 2 and #(r.actions & RemoteRead) = 1 and #(r.actions & ExternWrite) = 1}

fact {all disj r: RemoteRead, w: ExternWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rg) iff 
			w in coget[r]
}

fact {all disj r: RemoteRead, w: ExternWrite | 
		(~actions[r] = ~actions[w]  and ~actions[w] in Rg) iff 
			(rV[r] = wV[w] and w in po[r] and 
			w = eactions[r])
}


/* ======================= */
/* Remote Get Accumulate statement */
/* ======================= */

sig Rga extends Statement {}

fact { all r: Rga | #r.actions = 3 and #(r.actions & ExternWrite) = 1 and 
							#(r.actions & ExternRead) = 1 and 
							#(r.actions & RemoteReadWrite) = 1
}

fact {all disj r: ExternRead, w: ExternWrite, rw: RemoteReadWrite |
		(~actions[r] = ~actions[w] and ~actions[w] = ~actions[rw] and ~actions[w] in Rga) iff
			(wV[w] = wV[rw] and wV[rw] = rV[r].plus[rV[rw]] and 
				rw in po[r] and rw in corga1[r] and w in po[rw] and w in corga2[rw]  and 
				r in eactions[rw] and w in eactions[rw] and #eactions[rw] = 2)
}

//fact {all disj r1, r2: Reader, w: Writer, rw: (Reader & Writer) | 1 = 1}

/* ========================= */
/* Remote Compare and Swap statement */
/* ========================= */

sig Rcas extends Statement {}

fact {all r: Rcas | #r.actions = 4 and #(r.actions & Reader) = 3 and #(r.actions & Writer) = 2}



/* All actions corresponding to a statement have the same origin thread */
fact { all s: Statement| #(s.actions.o) = 1}

/* All ExternalAction are triggered by a Statement */
fact {Statement.actions = ExternalAction}

/* Statements cannot share actions */
fact {all disj s1, s2: Statement | #(s1.actions & s2.actions) = 0}



abstract sig Action extends Item {
	/* program order, consistency order */
	po : set Action, 

	/* destination and origin thread of the action */	
	d, o : one Thr
	
}

/* local reads and writes */
abstract sig LocalAction extends Action {
	poico: set Action
} 

fact { all l: LocalAction| o[l] = d[l]}

/* external reads and writes */
abstract sig ExternalAction extends Action {
	coexternalflush: set Flush
}


/* local and external reads */
sig Reader in Action {
	rl: one MemoryLocation,
	W : one Writer,
	rV: one (Int + Top),
	
	cofr: set Writer,
	cofr_prime: set Writer
}

fact {all r: Reader| d[r] = host[rl[r]] }

/* local and external writes */
sig Writer in Action {
	wl: one MemoryLocation,
	wV: one (Int + Top),

	corf: set Reader,
	cows: set Writer
}

fact {all w: Writer| d[w] = host[wl[w]]}

/* ============ */
/* atomic statements */
/* ============ */
sig Atomic in Action { }

/* =========== */
/* external actions  */
/* =========== */
abstract sig RemoteAction extends ExternalAction {
	eactions: set ExternalAction
}

abstract sig ExternAction extends ExternalAction {}

fact {all e : ExternAction | d[e] = o[e]}


/* ========= */
/* extern read */
/* ========= */
abstract sig ExternRead extends ExternAction {

	coput: lone RemoteWrite,
	corga1: lone RemoteReadWrite	
}

fact {all a: ExternRead |
				a in Reader and 
				not (a in Writer)}

sig AExternRead extends ExternRead {}
fact {all a: AExternRead | a in Atomic}


sig NExternRead extends ExternRead {}
fact {all a: NExternRead | not a in Atomic}

/* ======== */
/* remote read */
/* ======== */
abstract sig RemoteRead extends RemoteAction {
	coget: one ExternWrite
}

fact {all a: RemoteRead |
				a in Reader and 
				not (a in Writer)}

sig ARemoteRead extends RemoteRead {}
fact {all a: ARemoteRead | a in Atomic}

sig NRemoteRead extends RemoteRead {}
fact {all a: NRemoteRead | not a in Atomic}

/* ========= */
/* extern write */
/* ========= */
abstract sig ExternWrite extends ExternAction {
}

fact {all a: ExternWrite |
				a in Writer and 
				not (a in Reader)}

sig AExternWrite extends ExternWrite {}
fact {all a: AExternWrite | a in Atomic}

sig NExternWrite extends ExternWrite {}
fact {all a: NExternWrite | not a in Atomic}

/* ======== */
/* remote write */
/* ======== */
abstract sig RemoteWrite extends RemoteAction {}

fact {all a: RemoteWrite |
				a in Writer and 
				not (a in Reader)}

sig ARemoteWrite extends RemoteWrite {}
fact {all a: ARemoteWrite | a in Atomic}

sig NRemoteWrite extends RemoteWrite {}
fact {all a: NRemoteWrite | not a in Atomic}

/* ============= */
/* remote read - write */
/* ============= */

abstract sig RemoteReadWrite extends RemoteAction{
	corga2: one ExternWrite
}

fact {all a : RemoteReadWrite | 
			a in Writer and 
			a in Reader}

sig ARemoteReadWrite extends RemoteReadWrite {}
fact {all a: ARemoteReadWrite | a in Atomic}

sig NRemoteReadWrite extends RemoteReadWrite {}
fact {all a: NRemoteReadWrite | not a in Atomic}



/* =======  */
/* local actions */
/* =======  */

/* ====== */
/*local write */
/* ====== */
abstract sig Write extends LocalAction {}

fact {all a: Write | 
				a in Writer and 
				not (a in Reader)}

sig AWrite extends Write {}
fact {all a: AWrite | a in Atomic}

sig NWrite extends Write {}
fact {all a: NWrite | not a in Atomic}

/* ======= */
/* initial value */
/* ======= */
sig InitialValue extends AWrite {
	coinit: set Action-InitialValue
}

fact {all i: InitialValue, a: Action-InitialValue | 
				(a in coinit[i]) and
				(o[i] = o[a] implies a in po[i])
}

/* ====== */
/* local read */
/* ====== */
abstract sig Read extends LocalAction {
	reg: one Register
}

fact {all disj a, b: Read|
				not (reg[a] = reg[b])}

fact {all a: Read | 
				a in Reader and 
				not (a in Writer)}

sig ARead extends Read {}
fact {all a: ARead | a in Atomic}

sig NRead extends Read {}
fact {all a: NRead | not a in Atomic}

/* ========== */
/* flush statement */
/* ========== */
sig Flush extends Action {
	coflushlocal: set LocalAction,
	coflushexternal: set ExternalAction	
}

fact {all a: Flush | 
				not (a in Reader) and 
				not (a in Writer)}

/* ===== */
/*  RULES  */
/* ===== */

fact { po = ^po }

//fact { poco = ^(po + co)}



/* program order only between threads */
fact {all disj a, b: Action | 
			(b in po[a] or a in po[b]) iff o[a] = o[b] 
}

/* reader and reads from have the same variable */
fact {all r : Reader | 
			rl[r] = wl[W[r]] and d[r] = d[W[r]]
}

/* CO => poico*/
fact {all a: LocalAction, b: Action | 
					(b in po[a] iff b in poico[a])
}

/* F1 => coflushlocal */
fact {all l: LocalAction, f: Flush |
			(l in po[f] iff l in coflushlocal[f]) 
}

/* F2 => coexternflush */
fact {all f: Flush, e: RemoteAction |
			(f in po[e] and d[f] = d[e]) iff 
				(f in coexternalflush[e] and all a: eactions[e] | f in coexternalflush[a])
}

/* F3 => coflushexternal*/
fact {all f: Flush, e: RemoteAction|
			(e in po[f] and d[f] = d[e]) iff 
				(eactions[e] in coflushexternal[f] and e in coflushexternal[f])
}



/* R2 => corf */
fact {all r: Reader, w: Writer| 
			(w = W[r] and r in Atomic and W[r] in Atomic) iff 
				(r in corf[w])
}

/* cows */
fact {all w1, w2: Writer| 
			(not (w1 = w2) and wl[w1] = wl[w2] and w1 in Atomic and w2 in Atomic) iff 
			   (w1 in cows[w2] or w2 in cows[w1]) }

/* In-order Routing - Optional */
/*fact {all e1, e2: RemoteAction | 
		(d[e1] = d[e2] and e2 in po[e1]) implies
			 e2 in co[e1]
}*/


/* conflicts */
pred conflictFree[r: Reader] {
	(r in Atomic) and 
	all w: Writer |
		rl[r] = wl[w] implies 
			((co[w, r] or co[r, w])
			or w in Atomic)
}

pred wConflictFree[w: Writer] {
	w in Atomic and 
	all w1: Writer - w | 
		(wl[w] = wl[w1] implies 
			((co[w1, w] or co[w, w1])
			or w1 in Atomic))
}

/* C1 and C2 */
fact {all r: Reader |
			(conflictFree[r] and wConflictFree[W[r]]) implies
				rV[r] = wV[W[r]] else rV[r] = Top
}

/* Robustness */

/* First robustness criterion */
/*pred first_property[r: Reader, w: Writer] {
	rl[r] = wl[w] and
	(poco[w, r] and (not co[w, r])) or 
			(poco[r, w] and  (not co[r, w])) 
}*/

/* Second robustness criterion */
/*pred second_property[r: Reader, w2: Writer] {
    rl[r] = wl[w2] and
	((poco[W[r], w2] and not (co[W[r], w2]))
    or 
    (poco[w2, W[r]] and not (co[w2, W[r]])))
}*/

/* Robustness rule */
/*fact {(all r: Reader, w: Writer | not first_property[r, w] and not second_property[r, w])  implies 
												Robust.is_robust = True else
												Robust.is_robust = False
}*/

/* C1 and C2 */
fact {(all r: Reader |
			conflictFree[r] and wConflictFree[W[r]]) implies
				Robust.no_conflicts = True else Robust.no_conflicts = False
}

sig Pivots in Action {}

fact { #(Pivots & Reader)=1 and #(Pivots & Writer)=1 and #Pivots = 2}

/* transitive closure of all rules except R1 on the pivots */
pred co_prime[a, b: Action] {
	 b in ^(poico + coflushlocal + coexternalflush+ 
		coflushexternal + corf + cofr_prime + coput +
		coget + corga1 + corga2 + coinit)[a]
}

/* transitive closure of all rules */
pred co[a, b: Action] {
	b in ^(poico + coflushlocal + coexternalflush+ 
		coflushexternal + corf + cofr + coput +
		coget + corga1 + corga2 + coinit)[a]
}

pred R1_hypothesis_co[r: Action, w: Action] {
			r in Reader & Atomic and
			w in Writer & Atomic and
			wl[w] = rl[r] and co[W[r], w] and
			not (w = W[r])
}

pred R1_hypothesis_co_prime[r: Action, w: Action] {
			r in Reader & Atomic and
			w in Writer & Atomic and
			wl[w] = rl[r] and co_prime[W[r], w] and
			not (w = W[r])
}

/* R1 => cofr */
fact{all disj r, w: Action |
			R1_hypothesis_co[r, w] iff 
				w in cofr[r] 
}

/* (R1 - Pivot) => cofr_prime */
fact{all disj r, w: Action |
			(not ((r in (Reader & Pivots)) and (w in (Writer & Pivots)))) implies 
				(R1_hypothesis_co_prime[r, w] iff 
					w in cofr_prime[r])
}

fact{ R1_hypothesis_co[Pivots & Reader, Pivots & Writer] and not wV[Pivots & Writer] = wV[W[Pivots&Reader]]}


/* Do not generate Top write values */
fact {all w: Writer | not wV[w] = Top}

pred acyclic[r: Action -> Action] {
	no a: Action| a in a.r
}

/* Program order is acyclic */
fact { acyclic[po] }

/* co_prime is acyclic */
fact { no a: Action | co_prime[a, a]}

/* for test generation, co is cyclic */
fact { not (no a: Action | co[a, a])}

/* Generate initial values for all variables */
fact {MemoryLocation = wl[InitialValue] }

/* Properties of the test */
pred show { #Thr = 2 and 

					#Action = 6 and
					#(Action.o) > 1 and 
					//#Read > 0 and 
					#Rp > 0 and

					#Rcas = 0 and
					#Rga = 0 and 
					//Robust.is_robust = True and
					Robust.no_conflicts = True}

/* Exists at most one initial value for each variable */
fact { all disj i, j: InitialValue | not wl[i] = wl[j]}


/* Generate distinct write values for initial values and local writes */
fact { all disj w1, w2: Write| not (wV[w1] = wV[w2])}
fact { all disj w1, w2: Write| (co[w1, w2]) implies wV[w2] > wV[w1]}
fact { all w: Write| 0 <=  wV[w] and wV[w] < #Write}


/* Witness */

 sig Witness in Read {}

fact {#Witness = 1}

fact { #(Pivots & Read) = 1 iff #(Witness & Pivots) = 1 }

//TODO: works for Put and Get, does not work for Rga or Cas
fact { #(Pivots & Read) = 0 iff ~actions[W[Witness]] = ~actions[Pivots & Reader] }

//fact { all w: Writer| (w = W[Witness]) or (not wV[w] = wV[W[Witness]])}

run show for 6













































