open micro_rma

pred peterson [] {

/* T0:
  f1 = 0
  t = 0
  X = 1
  Z0 = 0

  RemotePut(f0, X)
  flush
  t = 1
  flush
  r lf1 = f1; 
  r lt0 = t; 
  if (lf1 != 1 or lt0 != 1) 
    --Critical Section--

spec: mutual exclusion */

some disj rf1, rt, rf0: AtomicRead, 
			disj wf1, wt0, wx, wz0, wf0, wy, wz1: InitialValue,
			disj wt1: AtomicWrite, 
			disj rpf0x, rpf1y, rptz : RemotePut,
			disj vrf0x, vrf1y, vrtz, vrt: AVirtualRead,
			disj vwf0x, vwf1y, vwtz, vwt: AVirtualWrite,
			disj rgt: RemoteGet,
			disj  flush3, flush4, flush5, flush6 :Flush, 
			disj  vflush3, vflush4, vflush5, vflush6 :VirtualFlush, 
			disj t1, t2: Thread, 
			disj f0, f1, t, x, z0, z1, y, v: MemoryLocation | 

	/* thread 1 */
	/* f1 = 0 */
	o[wf1] = t1 and d[wf1] = t1 and wl[wf1] = f1 and V[wf1] = 0 and
	/* t = 0 */
	o[wt0] = t1 and d[wt0] = t1 and wl[wt0] = t and V[wt0] = 0  and 
	/* x = 1 */
	o[wx] = t1 and d[wx] = t1 and wl[wx] = x and V[wx] = 1 and 
	/* z0 = 0 */
	o[wz0] = t1 and d[wz0] = t1 and wl[wz0] = z0 and V[wz0] = 0 and 

	/* RemotePut(f0, x) */
	o[rpf0x] = t1 and d[rpf0x] = t2 and 

	trigger[vrf0x] = rpf0x and rl[vrf0x] = x and d[vrf0x] = t1 and
	trigger[vwf0x] = rpf0x and wl[vwf0x] = f0 and d[vwf0x] = t2 and

	/* flush */
	o[flush5] = t1 and d[flush5] = t2 and flush5 in po[rpf0x] and triggerFlush[vflush5] = flush5 and
	/* t = 1 */
	o[wt1] = t1 and d[wt1] = t1 and wl[wt1] = t and V[wt1] = 1 and wt1 in po[flush5] and
	/* flush */
	o[flush3] = t1 and d[flush3] = t2 and flush3 in po[wt1] and triggerFlush[vflush3] = flush3 and
	/* lf1 = f1 */
	o[rf1] = t1 and d[rf1] = t1 and rl[rf1] = f1 and rf1 in po[flush3] and 
	/* lt0 = t */
	o[rt] = t1 and d[rt] = t1 and rl[rt] = t and rt in po[rf1] and 

/* T1:
  f0 = 0
  Y = 1
  Z1 = 0
  flush
  RemotePut( f1, Y)
  flush 
  RemotePut( t, Z) 
  flush 
  r lf0, f0;
  v = RemoteGet( t)
  if (lf0 != 1 or lt1 != 0) do 
    --Critical Section-- */

	/* f0 = 0 */
	o[wf0] = t2 and d[wf0] = t2 and wl[wf0] = f0 and V[wf0] = 0 and	
	/* y = 1 */
	o[wy] = t2 and d[wy] = t2 and wl[wy] = y and V[wy] = 1 and 
	/* z1 = 0 */
	o[wz1] = t2 and d[wz1] = t2 and wl[wz1] = z1 and V[wz1] = 0 and 	

	/* RemotePut(f1, Y) */
	o[rpf1y] = t2 and d[rpf1y] = t1 and 

	trigger[vrf1y] = rpf1y and rl[vrf1y] = y and d[vrf1y] = t2 and
	trigger[vwf1y] = rpf1y and wl[vwf1y] = f1 and d[vwf1y] = t1 and

	/* flush */
	o[flush6] = t2 and d[flush6] = t1 and flush6 in po[rpf1y] and
	/* RemotePut (t, z) */
	o[rptz] = t2 and d[rptz] = t1 and rptz in po[flush6] and triggerFlush[vflush6] = flush6 and

	trigger[vrtz] = rptz and rl[vrtz] = z1 and d[vrtz] = t2 and
	trigger[vwtz] = rptz and wl[vwtz] = t and d[vwtz] = t1 and


	/* flush */
	o[flush4] = t2 and d[flush4] = t1 and flush4 in po[rptz] and 
	triggerFlush[vflush4] = flush4 and
	/* lf0 = f0 */
	o[rf0] = t2 and d[rf0] = t2 and rl[rf0] = f0 and rf0 in po[flush4] and 
	/* v = RemoteGet(t) */
	o[rgt] = t2 and d[rgt] = t1 and rgt in po[rf0] and

	trigger[vrt] = rgt and rl[vrt] = t and d[vrt] = t1 and
	trigger[vwt] = rgt and wl[vwt] = v and d[vwt] = t2 and

	/* spec */
	// (((rV[rf1] = 0) or (rV[rt] = 0)) and ((rV[rf0] = 0) or (wV[vwt] = 1) )) and

	/* program size */
	#Item = 31

}

run peterson for 31
