abstract sig Action {
	po, hb, co : set Action, /* program order, happens before, consistency order */
	o, d : some Thread,
}

fact {
	no a: Action| a in a.^po
	all a, b: Action| b in a.po implies a.o = b.o
}

fact {
	no a: Action| a in a.co
}


/* === Memory action === */
abstract sig MA extends Action {
	rl, wl : some MemoryLocation, 
	V: one Int
}

sig Reader in MA {
	W: one Writer
}

/* defines the Memory Actions which read a value */
fact readers {
	Reader = Read + RemoteGet + RA
}

one sig InitWrite in MA {}




sig Writer in MA {}

/* defines the Memory Actions which write a value */
fact writers {
	Writer = Write + RemotePut + RA + InitWrite
}

abstract sig RPG extends MA {}
abstract sig RA extends MA {}
abstract sig VA extends MA {op : one RPG + RA}

/* local actions */
sig Write extends MA {}
sig Read extends MA {}

/* remote actions */
sig RemotePut extends RPG {}
sig RemoteGet extends RPG {}
sig RemoteGetAccumulate extends RA {}
sig RemoteCAS extends RA {}

/* virtual actions */
sig VirtualRead extends VA {} {op in RemoteGet}
sig VirtualWrite extends VA {} {op in RemotePut}
sig VirtualRGA extends VA {} {op in RemoteGetAccumulate}
sig VirtualCAS extends VA {} {op in RemoteCAS}

sig Thread {}
sig MemoryLocation {}


/* === Synchronization action === */
abstract sig SA extends Action {
	so : set SA /* total order */
}

sig Flush extends SA {}
sig ExternalSyncSend extends SA {}
sig ExternalSyncRecv extends SA {
	send: one ExternalSyncSend
}


fun between[a, b: Action]: set Action{
	{ x: Action| x in po[a] and b in po[x] }
}

pred parallel[a, b: Action, Z : Action -> Action] {
	not (a in Z[b] or b in Z[a])
}


/* === Valid Executions === */

/* hb is acyclic (excluding fences, unlocks and flushes) */
fact e {
	no a: MA + VA| a in a.^hb
}

/* === Epochs and sync semantics === */

/* hb (d, {r, w}) ⇒ co (d, {r, w}) (17) */
/* hb ({r, w}, d) ⇒ co ({r, w}, d)  (18) */
fact fg {
	all a: Write + Read, b: VA | 
					(a in b.hb implies a in b.co) and
					(b in a.hb implies b in a.co)
}

/* {ra, rg}.d = d.o ⇒ co (d, {ra, rg})  (19) */
fact h {
	all a: RA + RemoteGet, b: VA|
					a.d = b.o implies a in b.co
}

/* po ({ra, rp, r, w}, fl) ⇒ co ({ra, rp, r, w}, fl)  (20) */
fact i {
	all a: RA + RemotePut + Read + Write, b: Flush|
			b in a.po implies b in a.co
}


fact j1 {
	all a: Read + Write, b: Flush| 
			a in b.po implies a in b.co
}

fact j2 {
	all a: RA + RemotePut, b: Flush|
			a in b.po and a.d = b.d implies a in b.co
}

/* hb (d, {ca, d}) ⇒ co (d, {ca, d})  (23) */
fact l {
	all a: VA, b: RPG + RA + VA|
			b in a.hb implies b in a.co
}

/* po ({r, w}, ca) ⇒ co ({r, w}, ca)  (24) */
fact m {
	all a: Read + Write, b: RPG + RA |
			b in a.po implies b in a.co
}

/* === Rules for consistent memory operations === */

pred data_race[a, b: Action] {
	(not a in b.hb and not b in a.hb) or
	(not a in b.co and not b in a.co)
}

/* no data races */
fact no_data_races {
	no disj a, b: Action| data_race[a, b]
}

pred cohb[a, b: Action] {
	b in a.hb and b in a.co
}

/* cohb( W(r), r ) and V( r ) = V( W(r) )  (28) */
fact f28 {
	all r: Reader| cohb[W[r], r] and V[r] = V[W[r]]
}

/* cohb( W(r), w ) and cohb( w, r ) ⇒ r.rl != w.wl and V( r ) = V( W(r) )  (29) */
fact f29 {
	/* Que: can we eliminate V[r] = V[W[r]] here? it's implied by previous */
	all r: Reader, w: Writer|
			cohb[W[r], w] and cohb[w, r] implies r.rl != w.wl
}

/* not  cohb( r, W( r ) )  (30) */
fact f30 {
	no r: Reader| cohb[r, W[r]]
}


/* === Memory ordering rules === */
fact f31 {
	all a, b: RA| (a.wl = b.wl and b in po[a]) implies b in co[a]
}


pred program1 [] {
	some rx, ry: Read, wx, wy: RemotePut, t1, t2: Thread, x, y: MemoryLocation |
		t1 != t2 and x != y and 
		o[rx] = t1 and o[wy] = t1 and 
		o[ry] = t2 and o[wx] = t2 and

		d[wx] = t1 and d[wy] = t2 and

		rx in po[wy] and ry in po[wx]
}

run program1



pred show{ some VA and #Thread = 2 and #MemoryLocation = 2}

run show for 8

