open base_rules as b

/* Pivots for rule R1 */
one sig ReaderPivot in Reader{} 
one sig WriterPivot in Writer{}

fact { not ReaderPivot = WriterPivot }

/* consistency order without R1 on Pivots */
pred co_prime[a, b: Action, t: Time] {
	 b in ^(poico + coflushlocal + coexternalflush+ cocas1 + cocas2 + coir + 
		coflushexternal + corf + cofr_prime[t] + coput +
		coget + corga1 + corga2 + coinit)[a]
}

pred R1_hypothesis_co[r: Action, w: Action, t: Time] {
			(not (r = w)) and
			(r in (Reader & Atomic)) and
			(w in (Writer & Atomic)) and
			(wl[w] = rl[r]) and 
			co[W[r], w, t] and
			(not (w = W[r]))
}

pred R1_hypothesis_co_prime[r: Action, w: Action, t: Time] {
			(not (r = w)) and
			(r in (Reader & Atomic)) and
			(w in (Writer & Atomic)) and
			(wl[w] = rl[r]) and 
			co_prime[W[r], w, t] and
			(not (w = W[r]))
}

/* R1 => cofr */
fact{all disj r, w: Action, disj t, t': Time |
			mlt[t, t'] implies ((R1_hypothesis_co[r, w, t]) iff 
				w in cofr[t', r]) 
}

/* (R1 - Pivot) => cofr_prime */
fact{all disj r, w: Action, disj t, t': Time |
			(mlt[t, t']  and not ((r = ReaderPivot) and (w = WriterPivot)) ) implies 
				((R1_hypothesis_co_prime[r, w, t] ) iff 
					w in cofr_prime[t', r])
}

fact{R1_hypothesis_co[ReaderPivot, WriterPivot, mlast]}

/* The writer pivot writes different value from the W[the reader pivot] */
 fact{ not wV[WriterPivot] = wV[W[ReaderPivot]]}

/* Do not generate Top write values */
fact {all w: Write | not wV[w] = Top}

/* CO' is acyclic */
fact { no a: Action | co_prime[a, a, mlast]}

/* CO is cyclic, and the reader pivot is involved in the cycle */
fact {co[ReaderPivot, ReaderPivot, mlast]}

/* Witness */
one sig Witness in Read {}

/* Witness reads the same variable as the reader Pivot*/
fact{ rl[Witness] = rl[ReaderPivot]}

/* R2 rule (corf) */
fact {all disj r: Reader, w: Writer| 
			(r in Atomic and w in Atomic and w = W[r]) iff 
				(r in corf[w])
}

/* IR rule - Optional */
fact {all disj e1, e2: RemoteAction | 
		(d[e1] = d[e2] and e2 in po[e1]) iff
			 e2 in coir[e1]
}

/* F2 rule (coexternflush) */
fact {all f: Flush, e: RemoteAction |
			(f in po[e] and d[f] = d[e]) iff 
				(f in coexternalflush[e] and all a: eactions[e] | f in coexternalflush[a])
}
