// -*-  indent-tabs-mode:nil; c-basic-offset:4; -*-
import java.util.List;
import java.util.ArrayList;

public class DmappVisitor extends TLBaseVisitor<Object> {
    InstanceValue result;
    StringBuilder simpleExpressionBuffer, expressionBuffer;

    public static final String errorCheck =
            "if (status != DMAPP_RC_SUCCESS) {\n" +
            "\t\t\tdmapp_explain_error(status, (const char **)&expl);\n" +
            "\t\t\tfprintf(stderr, \" dmapp_get FAILED: w/ %d - %s -%s \\n\", status, dmapp_err_str[status], expl);\n" +
            "\t\t\texit(1);\n" +
            "\t\t}";

    Proc currentProc;
    int counter;

    public DmappVisitor(InstanceValue result) {
        this.result = result;
        simpleExpressionBuffer = new StringBuilder();
        expressionBuffer = new StringBuilder();
    }


    @Override
    public Object visitProcess(TLParser.ProcessContext ctx) {
        int procNumber = Integer.parseInt(ctx.Number().getText());
        this.currentProc = new Proc(procNumber);
        result.processes.add(this.currentProc);
        result.processContents.put(this.currentProc.procNumber, new ArrayList<>());

        visitChildren(ctx);
        return null;
    }

    @Override
    public Object visitSharedDecl(TLParser.SharedDeclContext ctx) {
        return null;
    }

    @Override
    public Object visitLocalDecl(TLParser.LocalDeclContext ctx) {
        return null;
    }

    @Override
    public Object visitStatement(TLParser.StatementContext ctx) {
        visitChildren(ctx);
        return null;
    }

    @Override
    public Object visitAssignment(TLParser.AssignmentContext ctx) {
        List<String> insts = result.processContents.get(this.currentProc.procNumber);

        String lhs = ctx.Identifier().getText();

        expressionBuffer.setLength(0);
        if (ctx.expression() instanceof TLParser.NumberExpressionContext)
            visitNumberExpression((TLParser.NumberExpressionContext)ctx.expression());
        else {
            System.out.println("unhandled expr in assignment stmt");
            System.exit(1);
        }

        String rhs = expressionBuffer.toString();

        String newInst = String.format("%s = %s;", lhs, rhs);
        insts.add(newInst);

        return null;
    }

    @Override
    public Object visitFlush(TLParser.FlushContext ctx) {
        List<String> insts = result.processContents.get(this.currentProc.procNumber);
        insts.add("dmapp_gsync_wait();");

        visitChildren(ctx);
        return null;
    }

    @Override
    public Object visitStore(TLParser.StoreContext ctx) {
        List<String> insts = result.processContents.get(this.currentProc.procNumber);

        String writeVar = ctx.Identifier().getText();
        String rhs = ctx.rhs().getText();
        String writeAtomicity = ctx.atomicity().getText();

        String newInst = String.format("*%s = %s;", writeVar, rhs);
        insts.add(newInst);

        return null;
    }

    @Override
    public Object visitLoad(TLParser.LoadContext ctx) {
        List<String> insts = result.processContents.get(this.currentProc.procNumber);

        String lhs = ctx.Identifier(0).getText();
        String readVar = ctx.Identifier(1).getText();
        String readAtomicity = ctx.atomicity().getText();

        String newInst = String.format("%s = *%s;", lhs, readVar);
        insts.add(newInst);

        return null;
    }

    @Override
    public Object visitGet(TLParser.GetContext ctx) {
        List<String> insts = result.processContents.get(this.currentProc.procNumber);

        int destProcessNumber = Integer.parseInt(ctx.Number().getText());
        String writeVar = ctx.Identifier(0).getText();
        String readVar = ctx.Identifier(1).getText();
        String readName = "vr" + counter;
        String writeName = "vw" + counter;

        String writeAtomicity = ctx.atomicity(0).getText();
        String readAtomicity = ctx.atomicity(1).getText();

        String newInst = String.format
            ("status = dmapp_get_nbi(%s, %s, seg, %d, 1, DMAPP_QW);",
             writeVar, readVar, destProcessNumber);
        insts.add(newInst);
        insts.add(errorCheck);

        return null;
    }

    @Override
    public Object visitPut(TLParser.PutContext ctx) {
        List<String> insts = result.processContents.get(this.currentProc.procNumber);

        int destProcessNumber = Integer.parseInt(ctx.Number().getText());
        String writeVar = ctx.Identifier(0).getText();
        String readVar = ctx.Identifier(1).getText();

        String writeAtomicity = ctx.atomicity(0).getText();
        String readAtomicity = ctx.atomicity(1).getText();

        String newInst = String.format
            ("status = dmapp_put_nbi(%s, seg, %d, %s, 1, DMAPP_QW);",
             writeVar, destProcessNumber, readVar);
        insts.add(newInst);
        insts.add(errorCheck);

        return null;
    }

    @Override
    public Object visitRhs(TLParser.RhsContext ctx) {
        if (ctx.Number() != null) {
            simpleExpressionBuffer.append(ctx.getText());
        } else {
            simpleExpressionBuffer.append(ctx.getText() + ".value");
        }
        return null;
    }

    @Override
    public Object visitNumberExpression(TLParser.NumberExpressionContext ctx) {
	expressionBuffer.append(ctx.Number());
	return null;
    }

    @Override
    public Object visitSimpleExpression(TLParser.SimpleExpressionContext ctx) {
        if (ctx.And() != null) {
            /* simpleExpression And simpleExpression*/
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(" && ");
            visitSimpleExpression(ctx.simpleExpression(1));
            simpleExpressionBuffer.append(")");

        } else if (ctx.Or() != null) {
            /* simpleExpression Or simpleExpression */
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(" || ");
            visitSimpleExpression(ctx.simpleExpression(1));
            simpleExpressionBuffer.append(")");

        } else if (ctx.Excl() != null) {
            /* Excl simpleExpression */
            simpleExpressionBuffer.append(" !(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(")");

        } else if (ctx.OParen() != null) {
            /* OParen simpleExpression CParen */
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(")");

        }else if (ctx.Equals() != null) {
            /* Identifier Equals rhs */
            simpleExpressionBuffer.append(ctx.Identifier().getText() + " == " );
            visitRhs(ctx.rhs());

        } else if (ctx.NEquals() != null) {
            /* Identifier NEquals rhs */
            simpleExpressionBuffer.append("(not (" + ctx.Identifier().getText() + " != " );
            visitRhs(ctx.rhs());
            simpleExpressionBuffer.append("))");

        } else {
            System.err.println("Error in visitSimpleExpression");
            System.exit(1);
        }
        return null;
    }

    @Override
    public Object visitAssumption(TLParser.AssumptionContext ctx) {
        /* Reset the buffer */
        simpleExpressionBuffer.setLength(0);

        visitSimpleExpression(ctx.simpleExpression());

        // System.out.println("Assumption found: " + simpleExpressionBuffer.toString());

        return null;
    }

    @Override
    public Object visitAssertion(TLParser.AssertionContext ctx) {
        /* Reset the buffer */
        simpleExpressionBuffer.setLength(0);

        visitSimpleExpression(ctx.simpleExpression());

        /* Append buffer to result */
        result.owningProcess.put(ctx.simpleExpression(), currentProc.procNumber);
        result.assertions.add(ctx.simpleExpression());

        // System.out.println("Assertion found: " + result.Assertion.toString());

        return null;
    }

    @Override
    public Object visitOutput(TLParser.OutputContext ctx) {
        simpleExpressionBuffer.setLength(0);

        visitSimpleExpression(ctx.simpleExpression());

        result.outputs.add(simpleExpressionBuffer.toString());
        return null;
    }

    @Override
    public Object visitRga(TLParser.RgaContext ctx) {
        int destProcessNumber = Integer.parseInt(ctx.Number().getText());
        String localWriteVar = ctx.Identifier(0).getText();
        String remoteWriteVar = ctx.Identifier(1).getText();
        String readVar = ctx.Identifier(2).getText();

        List<String> insts = result.processContents.get(this.currentProc.procNumber);

        String newInst = String.format
                ("status = dmapp_afadd_qw_nbi(%s, %s, seg, %d, *%s);",
                        localWriteVar, remoteWriteVar, destProcessNumber, readVar);
        insts.add(newInst);
        insts.add(errorCheck);

        return null;
    }

    @Override
    public Object visitCas(TLParser.CasContext ctx) {
        int destProcessNumber = Integer.parseInt(ctx.Number().getText());
        String writeVar = ctx.Identifier(0).getText();
        String swapVar = ctx.Identifier(1).getText();
        String readCompareVar = ctx.Identifier(2).getText();
        String readNewVar = ctx.Identifier(3).getText();

        String readCompareAtomicity = ctx.atomicity(0).getText();
        String readNewAtomicity = ctx.atomicity(1).getText();
        String swapAtomicity = ctx.atomicity(2).getText();
        String writeAtomicity = ctx.atomicity(3).getText();

        List<String> insts = result.processContents.get(this.currentProc.procNumber);

        String newInst = String.format
                ("status = dmapp_acswap_qw_nbi(%s, %s, seg, %d, *%s, *%s);",
                        writeVar, swapVar, destProcessNumber, readCompareVar, readNewVar);
        insts.add(newInst);
        insts.add(errorCheck);

        return null;
    }

    @Override
    public Object visitIfStatement(TLParser.IfStatementContext ctx) {
        simpleExpressionBuffer.setLength(0);

        visitSimpleExpression(ctx.simpleExpression());

        return null;
    }
}
