#include <stdio.h>
#include <unistd.h>
#include "pmi.h"
#include "dmapp.h"

/*
	T1:			T2:
	Y = 0			X = 0
	Z = 1			a = X
	RemotePut(X, Z)							
	Y = RemoteGet(X)
	flush
	Z = 5
	RemotePut(X, Z)

	spec: a is 0 or 1 or 5

*/
int main(int argc, char **argv) {
	char 			*expl = NULL;
	int 			pe = -1;
	int 			npes = -1;
	int 			target_pe = -1;
	long 			a = -1L;
	long 			*X = NULL;	
	long 			*Y = NULL;
	long			*Z = NULL;	
	dmapp_return_t 		status;
	dmapp_rma_attrs_t 	actual_args = {0}, rma_args = {0};
	dmapp_jobinfo_t 	job;
	dmapp_seg_desc_t 	*seg = NULL;
	
	/* Set the RMA parameters. */
	rma_args.put_relaxed_ordering = DMAPP_ROUTING_ADAPTIVE;
	rma_args.max_outstanding_nb = DMAPP_DEF_OUTSTANDING_NB;
	rma_args.offload_threshold = DMAPP_OFFLOAD_THRESHOLD;
	rma_args.max_concurrency = 1;
	
	/* Initialize DMAPP resources before executing any other DMAPP calls. */
	status = dmapp_init(&rma_args, &actual_args);
	if(status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_init returned %d - %s - %s \n", status, dmapp_err_str[status], expl);
		exit(1);
	}
	
	/* Retrieve information about job details such as PE id and number of PEs. */
	status = dmapp_get_jobinfo(&job);
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_get_jobinfo FAILED: w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	
	pe = job.pe;
	npes = job.npes;	
	target_pe = npes - pe - 1;

	/* Specify in which segment the remote memory region (the source) lies. 
	   In this case, it is the heap (see above). */
  	seg = &(job.sheap_seg);
	
	/* Allocate remotely accessible memory for the source and target bufferes. 
	   Only memory in the data segment or the sheap is remotely accessible. 
	   Here we allocate from the sheap. */
	X = (long *)dmapp_sheap_malloc(sizeof(long));
	Y = (long *)dmapp_sheap_malloc(sizeof(long));
	Z = (long *)dmapp_sheap_malloc(sizeof(long));
	if (X == NULL || Y == NULL || Z == NULL) {
		fprintf(stderr, " sheap_malloc'd failed X 0x%lx Y 0x%lx Z 0x%lx\n", (long)X, (long)Y, (long)Z);	
		exit(1);
	}
	
	/* Synchronize to make sure everyone's buffers are initialized before data transfer is started.  */ 
	PMI_Barrier();	
	
	/* Initialization (T1 == even PEs; T2 == odd PEs). */
	if((pe%2) == 0) {
		*Y = 0L;
		*Z = 1L;
		status = dmapp_put_nbi(X, seg, target_pe, Z, 1, DMAPP_QW);
		if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_put FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
		status = dmapp_get_nbi(Y, X, seg, target_pe, 1, DMAPP_QW);
		if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
		dmapp_gsync_wait();
		*Z = 5L;
		sleep(5);
		status = dmapp_put_nbi(X, seg, target_pe, Z, 1, DMAPP_QW);
		if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_put FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
	} else {
		*X = 0L;
		sleep(1);
		a = *X;
	}
	
	/* Synchronize before verifying the data. */
	PMI_Barrier();

	/* Verify the data. */
	if (((pe%2) != 0) && (a != 0L && a != 1L && a != 5L)) {
		fprintf(stderr, " test10 FAILED for PEs (%d,%d): a is %ld, should be 0 or 1 or 5\n", pe, target_pe, a);
	}

	if(pe%2 != 0) fprintf(stderr, "(%ld)", a);

	/* Free buffers allocated from sheap. */
	dmapp_sheap_free(X);
	dmapp_sheap_free(Y);
	
	/* Release DMAPP resources. This is a mandatory call. */
	status = dmapp_finalize();
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_finalize FAILED w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	

	return(0);
}

