#include <stdio.h>
#include <unistd.h>
#include "pmi.h"
#include "dmapp.h"

/*
	P1:			P2:
	X = remoteGet(X)	X = 0
	X = 1
	a = X

	spec: (a == 1L) || (a == 0L)

*/
int main(int argc, char **argv) {
	char 			*expl = NULL;
	int 			pe = -1;
	int 			npes = -1;
	int 			target_pe = -1;
	long 			a = -1L;
	long 			*X = NULL;	
	long 			*Y = NULL;
	dmapp_return_t 		status;
	dmapp_rma_attrs_t 	actual_args = {0}, rma_args = {0};
	dmapp_jobinfo_t 	job;
	dmapp_seg_desc_t 	*seg = NULL;
	
	/* Set the RMA parameters. */
	rma_args.put_relaxed_ordering = DMAPP_ROUTING_ADAPTIVE;
	rma_args.max_outstanding_nb = DMAPP_DEF_OUTSTANDING_NB;
	rma_args.offload_threshold = DMAPP_OFFLOAD_THRESHOLD;
	rma_args.max_concurrency = 1;
	
	/* Initialize DMAPP resources before executing any other DMAPP calls. */
	status = dmapp_init(&rma_args, &actual_args);
	if(status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_init returned %d - %s - %s \n", status, dmapp_err_str[status], expl);
		exit(1);
	}
	
	/* Retrieve information about job details such as PE id and number of PEs. */
	status = dmapp_get_jobinfo(&job);
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_get_jobinfo FAILED: w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	
	pe = job.pe;
	npes = job.npes;	
	target_pe = npes - pe - 1;

	/* Specify in which segment the remote memory region (the source) lies. 
	   In this case, it is the heap (see above). */
  	seg = &(job.sheap_seg);
	
	/* Allocate remotely accessible memory for the source and target bufferes. 
	   Only memory in the data segment or the sheap is remotely accessible. 
	   Here we allocate from the sheap. */
	X = (long *)dmapp_sheap_malloc(sizeof(long));
	Y = (long *)dmapp_sheap_malloc(sizeof(long));
	if (X == NULL || Y == NULL) {
		fprintf(stderr, " sheap_malloc'd failed X 0x%lx Y 0x%lx\n", (long)X, (long)Y);	
		exit(1);
	}
	
	/* Synchronize to make sure everyone's buffers are initialized before data transfer is started.  */ 
	PMI_Barrier();	
	
	/* Initialization (T1 == even PEs; T2 == odd PEs). */
	if((pe%2) == 0) {
		status = dmapp_get_nbi(X, X, seg, target_pe, 1, DMAPP_QW);
		if (status != DMAPP_RC_SUCCESS) {
			dmapp_explain_error(status, (const char **)&expl);
			fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
			exit(1);
		}
		*X = 1L;
		a = *X;
	} else {
		*X = 0L;
	}
	
	/* Synchronize before verifying the data. */
	PMI_Barrier();

	/* Verify the data. */
	if (((pe%2) == 0) && !((a == 1L) || (a == 0L))) {
		fprintf(stderr, " FAILED:(%ld);", a);
	}
		
	/* Free buffers allocated from sheap. */
	dmapp_sheap_free(X);
	dmapp_sheap_free(Y);
	
	/* Release DMAPP resources. This is a mandatory call. */
	status = dmapp_finalize();
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_finalize FAILED w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	

	return(0);
}

