#include "rdma-common.h"

const int TIMEOUT_IN_MS = 500; /* ms */
const int NUM_ITERATIONS = 1000;

static int on_addr_resolved(struct rdma_cm_id *id);
static int on_connection(struct rdma_cm_id *id);
static int on_disconnect(struct rdma_cm_id *id);
static int on_event(struct rdma_cm_event *event);
static int on_route_resolved(struct rdma_cm_id *id);
static void usage(const char *argv0);

int main(int argc, char **argv) {
	int i;
	for(i=0; i<NUM_ITERATIONS; i++) {
		struct addrinfo *addr;
		struct rdma_cm_event *event = NULL;
		struct rdma_cm_id *conn= NULL;
		struct rdma_event_channel *ec = NULL;
	
		if (argc != 4)
			usage(argv[0]);
	
		set_app(CLIENT);
	
		int num;
		sscanf (argv[1],"%d",&num);
		set_test_num(num);
	
		TEST_NZ(getaddrinfo(argv[2], argv[3], NULL, &addr));
		
		TEST_Z(ec = rdma_create_event_channel());
		
		TEST_NZ(rdma_create_id(ec, &conn, NULL, RDMA_PS_TCP));
		TEST_NZ(rdma_resolve_addr(conn, NULL, addr->ai_addr, TIMEOUT_IN_MS));
	
		freeaddrinfo(addr);
	
		while (rdma_get_cm_event(ec, &event) == 0) {
			struct rdma_cm_event event_copy;
	
			memcpy(&event_copy, event, sizeof(*event));
			rdma_ack_cm_event(event);
	
			if (on_event(&event_copy))
				break;
			}
		
		rdma_destroy_event_channel(ec);
	}

	return 0;
}

int on_addr_resolved(struct rdma_cm_id *id) {
  build_connection(id);
  TEST_NZ(rdma_resolve_route(id, TIMEOUT_IN_MS));

  return 0;
}

int on_connection(struct rdma_cm_id *id) {
  on_connect(id->context);
  send_mr_X(id->context);

  return 0;
}

int on_disconnect(struct rdma_cm_id *id) {
  destroy_connection(id->context);
  return 1; /* exit event loop */
}

int on_event(struct rdma_cm_event *event) {
  int r = 0;

  if (event->event == RDMA_CM_EVENT_ADDR_RESOLVED)
    r = on_addr_resolved(event->id);
  else if (event->event == RDMA_CM_EVENT_ROUTE_RESOLVED)
    r = on_route_resolved(event->id);
  else if (event->event == RDMA_CM_EVENT_ESTABLISHED)
    r = on_connection(event->id);
  else if (event->event == RDMA_CM_EVENT_DISCONNECTED)
    r = on_disconnect(event->id);
  else
    die("on_event: unknown event.");

  return r;
}

int on_route_resolved(struct rdma_cm_id *id) {
  struct rdma_conn_param cm_params;

  build_params(&cm_params);
  TEST_NZ(rdma_connect(id, &cm_params));

  return 0;
}

void usage(const char *argv0) {
  fprintf(stderr, "usage: %s <test_num> <server-address> <server-port>\n", argv0);
  exit(1);
}
