open util/integer

one sig Top {}

/* Threads */

abstract sig Thr {}

sig NIC extends Thr {}

sig Thread extends Thr {
	nic: one NIC
}

fact { all disj t1, t2: Thread | nic[t1] != nic[t2]}



sig MemoryLocation {}

sig Item {}

abstract sig Action extends Item {
	/* program order, happens before, consistency order and sync order */
	po, hb, co, so, conflict : set Action, 
	/* destination thread of the action */	
	d : one Thread,	
	/* origin thread */
	o : one Thr,
	no_conflict: lone Action 
}

fact {all disj a, b: Action| not b in no_conflict[a]}

fact{all a:Action| not a in conflict[a]}

sig Atomic in Action {}

/* Types of actions */

abstract sig VirtualAction extends Action {
	trigger : one Trigger
} {
	o in NIC
}

abstract sig LocalAction extends Action {} {
	o in Thread
}


/* the action that triggers virtual actions */
abstract sig Trigger extends Action {
	virtuals: set VirtualAction
} {
	o in Thread
}

fact {
	all t: Trigger |
	not (t in Atomic) and
	not (t in Reader) and
	not (t in Writer)
}

sig Reader in Action {
	rl: one MemoryLocation,
	rV: one (Int + Top),
	W : one Writer,
	nonDeterministic: lone Reader
}

fact {all disj r1, r2: Reader| not r2 in nonDeterministic[r1]}

sig Writer in Action {
	wl: one MemoryLocation,

	/* the value in the memory location (may differ from V, due to conflicts) */
	wV: one (Int + Top),

	/* the value wanted to be written */ 
	V : one (Int + Top)
}

/* ==================== */

/* ================  */
/* non-atomic virtual actions */
/* ================  */

abstract sig VirtualRead extends VirtualAction {}
fact {all a: VirtualRead |
				a in Reader and 
				not (a in Writer)}

abstract sig VirtualWrite extends VirtualAction {}
fact {all a: VirtualWrite |
				a in Writer and 
				not (a in Reader)}


sig VirtualReadWrite extends VirtualAction {}
fact {all a: VirtualReadWrite |
				a in Reader and 
				a in Writer and 
				not (a in Atomic)}


/* ============== */
/* atomic virtual actions */
/* =============  */

sig NVirtualRead extends VirtualRead {}
fact {all a: NVirtualRead | 
				not (a in Atomic)}

sig AVirtualRead extends VirtualRead {}
fact {all a: AVirtualRead | 
				a in Atomic}

sig AVirtualWrite extends VirtualWrite {}
fact {all a: AVirtualWrite | 
				a in Atomic }

sig NVirtualWrite extends VirtualWrite {}
fact {all a: NVirtualWrite | 
				not (a in Atomic)}


sig AtomicVirtualReadWrite extends VirtualAction {}
fact {all a: AtomicVirtualReadWrite | 
				a in Atomic and 
				a in Reader and 
				a in Writer}


/* ======================  */
/* atomic and non-atomic local actions */
/* ======================  */

sig Write extends LocalAction {}
fact {all a: Write | 
				a in Writer and 
				not (a in Atomic) and 
				not (a in Reader)}

sig AtomicWrite extends LocalAction {}
fact {all a: AtomicWrite | 
				a in Atomic and 
				a in Writer and 
				not (a in Reader)}

sig InitialValue extends LocalAction {}
fact {all a: InitialValue | 
				a in Atomic and 
				a in Writer and 
				not (a in Reader) }


fact {all i: InitialValue, a: Action-InitialValue | 
				(o[i] = o[a]) implies a in po[i]}

fact {all i: InitialValue, a: Action-InitialValue | 
				a in co[i] and a in so[i]}

fact {all i: InitialValue, a: Action |
				not i in hb[a] and not i in co[a]}


sig Read extends LocalAction {}
fact {all a: Read | 
				a in Reader and 
				not (a in Atomic) and 
				not (a in Writer)}

sig AtomicRead extends LocalAction {}
fact {all a: AtomicRead | 
				a in Atomic and 
				a in Reader and 
				not (a in Writer)}


/* =========  */
/* trigger actions */
/* =========  */

sig RemotePut extends Trigger {}

fun virtualActions[t: Trigger]: set VirtualAction {
	~trigger[t]
}

fun typedVirtualActions[t: Trigger, S: set VirtualAction]: set VirtualAction {
	virtualActions[t] & S
}

fun myPlus[a, b: one (Int + Top)] : one (Int + Top) {
	(a = Top or b = Top) implies Top else a.plus[b]
}

fact { all va: VirtualAction | o[va] = nic[o[trigger[va]]]}

fact { all rp: RemotePut | 
					let va = virtualActions[rp] | 
					let vr = typedVirtualActions[rp, VirtualRead] | 
					let vw = typedVirtualActions[rp, VirtualWrite] | 

					#va = 2 and 

					#vw = 1 and 
					vw in co[vr] and
					vw in so[vr] and

					#vr = 1  and
					vr in co[rp] and
					vr in so[rp] and

					rV[vr] = V[vw]}


sig RemoteGet extends Trigger {}
fact { all rp: RemoteGet | 
					let va = virtualActions[rp] | 
					let vr = typedVirtualActions[rp, VirtualRead] | 
					let vw = typedVirtualActions[rp, VirtualWrite] | 

					#va = 2 and 

					#vw = 1 and 
					vw in co[vr] and
					vw in so[vr] and
					
					#vr = 1 and
					vr in co[rp] and
					vr in so[rp] and

					rV[vr] = V[vw]}


sig RemoteGetAccumulate extends Trigger {}
fact { all rp: RemoteGetAccumulate | 
					let va = virtualActions[rp] | 
					let vr = typedVirtualActions[rp, VirtualRead] | 
					let vw = typedVirtualActions[rp, VirtualWrite] | 
					let vrw = typedVirtualActions[rp, VirtualReadWrite] | 

					#va = 3 and 

					#(vw) = 1 and 
					vw in co[vrw] and
					vw in so[vrw] and

					#vrw = 1 and
 					vrw in co[vr] and
					vrw in so[vr] and

					#vr = 1 and
					vr in co[rp] and
					vr in so[rp] and

					V[vrw] = V[vw] and
					V[vrw] = myPlus[rV[vr], rV[vrw]]}


sig CAS extends Trigger {}
fact { all rp: CAS | 
					let va = virtualActions[rp] | 
					let vr = typedVirtualActions[rp, VirtualRead] | 
					let vw = typedVirtualActions[rp, VirtualWrite] | 
					let vrw = typedVirtualActions[rp, VirtualReadWrite] | 

					#va = 4 and 

					#vw = 1 and 
					vw in co[vrw] and
					vw in so[vrw] and


					#vrw = 1 and
 					(vr in ~co[vrw]) and
					(vr in ~so[vrw]) and


					#vr = 2 and
					vr in co[rp] and
					vr in so[rp]}

sig Flush extends Action {} {
	o in Thread
}

sig VirtualFlush extends Action {
	triggerFlush: one Flush
}
fact {
	all vf: VirtualFlush |
		not (vf in Reader) and
		not (vf in Writer) and
		not (vf in Atomic)
}

/* === Flush rules === */

fact {
	all vf: VirtualFlush | 
		d[vf] = d[triggerFlush[vf]] and 
		o[vf] = d[triggerFlush[vf]] and 
		triggerFlush[vf] in co[vf] and 
		triggerFlush[vf] in so[vf]
}

fact {
	all vf: VirtualFlush, a: Flush + Trigger + LocalAction | 
		d[vf] = o[a] implies 
			(vf in po[a] or a in po[vf])
}

fact {
	all vf: VirtualFlush, v: VirtualAction |
		((d[vf] = d[v] or o[triggerFlush[vf]] = d[v]) and 
		 triggerFlush[vf] in po[trigger[v]]) implies 
				(vf in co[v] and vf in so[v])
}

fact {
	all vf: VirtualFlush, a: Action |
		triggerFlush[vf] in po[a] implies vf in co[a] and vf in so[a]
}

fact { all f: Flush |
				not (f in Atomic) and
				not (f in Reader) and
				not (f in Writer)
}

/* =========== */
/* === RULES === */
/* =========== */

fact { 	po = ^po }

fact {	co = ^co }

 /* no cycles in program order */
fact { all a: Action | 
		not a in po[a] 
}


 /* consistency order is not reflexive */
fact {	all a: Action | 
		not a in co[a]
}

/* program order only between threads */
fact { 	all a, b: Action | 
		b in po[a] implies o[a] = o[b] 
}

/* so relation only appears inter threads */
/*fact { 	all a, b: Action | 
		b in so[a] implies o[a] != o[b]
}*/

/* hb is the transitive closure of (po U so U rf) */
fact { hb = ^(po + so + ~W) } //ok


fact { all r : Reader | 
		rl[r] = wl[W[r]] and 
		d[r] = d[W[r]] 			/* NEWLY ADDED */
} // ok

/* hb is not reflexive */
fact e {	all a: Action | 
				not a in hb[a] 
}

/* hb (d, {r, w}) ⇒ co (d, {r, w}) (17) */
/* hb ({r, w}, d) ⇒ co ({r, w}, d)  (18) */
fact fg {
	all a: LocalAction, b: VirtualAction | 
					(a in hb[b] implies a in co[b]) and
					(b in hb[a] implies b in co[a])
} //ok


/* po ({ra, rp, r, w}, fl) ⇒ co ({ra, rp, r, w}, fl)  (20) */
fact i {
	all a: Trigger + LocalAction, b: Flush |
			a in po[b] implies a in co[b]
} //ok

fact j1 {
	all a: LocalAction, b: Flush | 
			b in po[a] implies b in co[a]
} //ok


fact j2 {
	all a: Trigger, b: Flush |
			b in po[a] and d[a] = d[b] 
				implies b in co[a]
} //ok 
	//TODO: check if we can remove d[a] = d[b], since we have total flushes

fact {
	all a: VirtualAction, b: Flush |
		b in po[trigger[a]] implies b in co[a] and b in so[a]
}


/* po ({r, w}, ca) ⇒ co ({r, w}, ca)  (24) */
fact m {
	all a: LocalAction, b: Trigger |
			b in po[a] 
				implies b in co[a]
} //ok


pred cohb[a, b: Action] {
	b in hb[a] and b in co[a]
}


/* cohb( W(r), w ) and cohb( w, r ) ⇒ r.rl != w.wl and V( r ) = V( W(r) )  (29) */
fact f29 {
	all r: Reader, w: Writer |
			cohb[W[r], w] and cohb[w, r] 
				implies rl[r] != wl[w]
} //ok


/* not  cohb( r, W( r ) )  (30) */
fact f30 {
	no r: Reader | 
		cohb[r, W[r]]
}



/* ================== */
/* === Conflict definitions === */
/* ================== */

pred weak_conflicting[a, b: Action] {
	(not cohb[a, b] and not cohb[b, a]) and
	((a in Reader and b in Writer and rl[a] = wl[b]) or
	 (a in Writer and b in Reader and rl[b] = wl[a]) or
	 (a in Writer and b in Writer and wl[a] = wl[b]))
}

//Determinize conflicts
//fact { all disj a, b: (Writer & Atomic)| 
//	(wl[a] = wl[b] and d[a] = d[b]) implies 
//		((a in co[b] and a in so[b]) or (b in co[a] and b in so[a]))	}

fact {all disj a, b: Reader | rl[a] = rl[b] and b in hb[a] implies 
				(W[b] in hb[a] or W[b] = W[a])}

fact {all disj a, b: Reader | (rl[a] = rl[b] and not (b in hb[a]) and not (a in hb[b])) implies 
				W[b] = W[a]}

pred strong_conflicting[a, b: Action] {
	weak_conflicting[a, b] and ((not a in Atomic) or (not b in Atomic))
}

fact { all disj a, b: Action | a in conflict[b] iff strong_conflicting[a, b] }

fact { all a: Action | a in no_conflict[a] iff (a in Atomic and weak_conflict_free[a] and not a in InitialValue)}


pred write_conflict_free[a: Action] {
	all b: Writer | b != a implies 
		not strong_conflicting[a, b]
}

pred conflict_free[a: Action] {
	all b: Action | b != a implies 
		not strong_conflicting[a, b]
}

pred weak_conflict_free[a: Action] {
	all b: Action | b != a implies 
		not weak_conflicting[a, b]
}

fact {
	all a: Reader | conflict_free[a] implies 
		rV[a] = wV[W[a]] else rV[a] = Top
}

fact {
	all a: Writer | write_conflict_free[a] implies 
		wV[a] = V[a] else wV[a] = Top
}

/* Deterministic Network Routing - Optional */
/*fact {
	all a, b: VirtualAction | 
		(d[a] = d[b] and trigger[b] in po[trigger[a]]) 
			implies b in co[a] and b in so[a]
}*/

/* Indicate if execution is deterministic or not */
/* -  check if a read has several options to read from */
fact {
	all disj w1, w2: Writer, r: Reader |
		W[r] = w1 and 
		wl[w1] = wl[w2] and
		(not cohb[w1, w2] and not cohb[w2, w1]) and 
		(not wV[w1] = wV[w2])
		implies r in nonDeterministic[r]
}


pred show {#RemotePut = 2}//#Flush = 1 and #RemotePut = 1 and  #Thread = 3}

run show for 12













































