// VPI_verbs test code
// inspired by https://thegeekinthecorner.wordpress.com/2010/09/28/rdma-read-and-write-with-ib-verbs/
#include "rdma-test.h"

// shared var decls
long unexpected_cnt = 0;
long long * all_failed;
long long * client_state, * server_state;
long long * X;
long long * Y;
long long * c_sender;
long long * c_global;
long long * c_receiver;

// local decls
long long c;

void register_shared_vars(struct connection * conn) {
  posix_memalign((void**)&(conn->rdma_region), 64, sizeof(long long)*8);
  memset(conn->rdma_region, 0, sizeof(long long)*8);

  vars = conn->rdma_region;
  // initialize vars
  all_failed = &vars[0];
  if (app == CLIENT) {
    client_state = &vars[1];
    *client_state = 0;
    server_state = &vars[2];
  } else {
    client_state = &vars[1];
    server_state = &vars[2];
    *server_state = 0;
  }
  X = &vars[3];
  Y = &vars[4];
  c_global = &vars[7];
  c_receiver = &vars[6];
  c_sender = &vars[5];

  test_init(conn);

  TEST_Z(conn->rdma_mr = ibv_reg_mr(
				    s_ctx->pd,
				    conn->rdma_region,
				    sizeof(long long) * 8,
				    IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ | IBV_ACCESS_REMOTE_ATOMIC));
}

/* ****************************** START OF TESTS SECTION ****************************** */

/*
process 0 (client)
shared X = 0;

process 1 (server)
shared Y = 1;
local c;
Y = get(atm,atm)(X,0);
put(atm,atm)(X,0,Y);
flush (0);
load(atm) c = Y;
// ---
*/

void test_init(struct connection *conn) {
  if (app == CLIENT) {
    *X = 0;
    *all_failed = 1;
  }
  else {
    *Y = 1;
  }
}

void test(struct ibv_cq *cq, struct connection *conn) {
  // pre-test barrier
  if (app == CLIENT) {
    *client_state = 1;
    rdma_operation(conn, conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
    consume_one_cqe(cq);
    while (*server_state < 1) /* busy-wait */;
  } else {
    *server_state = 1;
    rdma_operation(conn, conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
    consume_one_cqe(cq);
    while (*client_state < 1) /* busy-wait */;
  }
  // test code
  if (app == CLIENT) {
    if (rand() % 2 == 1) { usleep(5 + (rand()%10)); }
  } else {
    if (rand() % 2 == 1) { usleep(1); }
    rdma_operation(conn, conn->peer_mr, X - vars, Y, conn->rdma_mr, IBV_WR_RDMA_READ, 0);
    if (rand() % 2 == 1) { usleep(1); }
    rdma_operation(conn, conn->peer_mr, X - vars, Y, conn->rdma_mr, IBV_WR_RDMA_WRITE, 0);
    if (rand() % 2 == 1) { usleep(1); }
    send_flush(cq, conn);
    if (rand() % 2 == 1) { usleep(1); }
    c = *Y;
  }
  // ---
  // post-test barrier
  if (app == CLIENT) {
    *client_state = 2;
    rdma_operation(conn, conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
    consume_one_cqe(cq);
    while (*server_state < 2) /* busy-wait */;
  } else {
    *server_state = 2;
    rdma_operation(conn, conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
    consume_one_cqe(cq);
    while (*client_state < 2) /* busy-wait */;
  }
}

void test_result(struct ibv_cq *cq, struct connection *conn) {
  if (app == SERVER) {
    *c_global = c;
    rdma_operation(conn, conn->peer_mr, c_global - vars, c_global, conn->rdma_mr, IBV_WR_RDMA_WRITE, 0);
  }
  send_flush(cq, conn);
#if DUMP_VARS
  if (app == CLIENT) {
    printf("values: X=%lld\n", *X);
  }
  if (app != CLIENT) {
    printf("values:  Y=%lld c_sender=%lld c_receiver=%lld c_global=%lld\n", *Y, *c_sender, *c_receiver, *c_global);
  }
#endif
  if (app == CLIENT) {
    *client_state = 3;
    rdma_operation(conn, conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
    consume_one_cqe(cq);
    while (*server_state < 3) /* busy-wait */;
  } else {
    *server_state = 3;
    rdma_operation(conn, conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
    consume_one_cqe(cq);
    while (*client_state < 3) /* busy-wait */;
  }
}

void collate_results(struct ibv_cq *cq, struct connection *conn) {

    /* Load results from global _received variables to locals */
    if (app == CLIENT) {
        c = *c_global;
    }

    /* Check which output is observed */
    if (app == CLIENT) {
        if (0) ;
        else {
          fprintf(stderr, "UNEXPECTED: c=%lld\n", c);
          unexpected_cnt++;
        }
    }

  // synch before disconnecting
  if (app == CLIENT) {
    *client_state = 4;
    rdma_operation(conn, conn->peer_mr, client_state - vars, client_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
    consume_one_cqe(cq);
    while (*server_state < 4) /* busy-wait */;
  } else {
    *server_state = 4;
    rdma_operation(conn, conn->peer_mr, server_state - vars, server_state, conn->rdma_mr, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED);
    consume_one_cqe(cq);
    while (*client_state < 4) /* busy-wait */;
  }
}

void print_results() {
    if (app == CLIENT) {
        fprintf(stderr, "unexpected count %ld\n", unexpected_cnt);
    }
}

/* ****************************** END OF TESTS SECTION ****************************** */
