// -*-  indent-tabs-mode:nil; c-basic-offset:4; -*-
package ch.ethz.srl;

import edu.mit.csail.sdg.alloy4compiler.ast.Sig;
import edu.mit.csail.sdg.alloy4compiler.ast.Sig.Field;
import edu.mit.csail.sdg.alloy4compiler.translator.A4Solution;
import edu.mit.csail.sdg.alloy4compiler.translator.A4Tuple;
import edu.mit.csail.sdg.alloy4compiler.translator.A4TupleSet;

import java.util.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import java.io.PrintWriter;

import ch.ethz.srl.util.PseudoTopologicalOrderer;

class TGMemoryLocation {
    A4CodeGen.State s;
    String label; // alloy name
    String prettyLabel;
    Integer initialValue;

    public TGMemoryLocation(A4CodeGen.State s,
                            String label) {
        this.s = s;
        this.label = label;
    }

    @Override
    public String toString() {
        if (prettyLabel == null) {
            if (s.memLocCount < SHARED_VAR_NAMES.length())
                prettyLabel = Character.toString(SHARED_VAR_NAMES.charAt(s.memLocCount++));

            if (prettyLabel == null)
                prettyLabel = label;
        }
        return prettyLabel;
    }

    public static final String SHARED_VAR_NAMES = "XYZWVU";
}

class TGRegister {
    A4CodeGen.State s;
    String label; // alloy name
    String prettyLabel;
    Integer initialValue;

    public TGRegister(A4CodeGen.State s,
                      String label) {
        this.s = s;
        this.label = label;
    }

    @Override
    public String toString() {
        if (prettyLabel == null) {
            if (s.registerCount < LOCAL_VAR_NAMES.length())
                prettyLabel = Character.toString(LOCAL_VAR_NAMES.charAt(s.registerCount++));

            if (prettyLabel == null)
                prettyLabel = label;
        }
        return prettyLabel;
    }

    public static final String LOCAL_VAR_NAMES = "abcdef";
}

interface TGWriter extends TGAction {
    public int getWV();
    public void setWV(int wV);
    public TGMemoryLocation getWl();
    public void setWl(TGMemoryLocation wl);
    public boolean isLocalWrite();
}

class TGWriterImpl extends TGActionImpl implements TGWriter {
    private TGMemoryLocation wl;
    private boolean isLocalWrite;
    int wV;

    public TGWriterImpl(A4CodeGen.State s,
                    String label, TGMemoryLocation wl, int wV) {
        super(s, label); setWl(wl); this.wV = wV;
    }

    public TGWriterImpl(A4CodeGen.State s,
                    String label, TGMemoryLocation wl, int wV, boolean isLocalWrite) {
        this(s, label, wl, wV);
        this.isLocalWrite = isLocalWrite;
    }

    @Override public String toString() {
        return String.format("store(atm) %s = %d", wl, wV);
    }

    public int getWV() { return wV; }
    public void setWV(int wV) {
        this.wV = wV;
    }

    public TGMemoryLocation getWl() { return wl; }
    public void setWl(TGMemoryLocation wl) {
        this.wl = wl;
    }

    public boolean isLocalWrite() { return isLocalWrite; }
}

class TGInitialValue extends TGWriterImpl {
    public TGInitialValue(A4CodeGen.State s,
                          String label, TGMemoryLocation wl, int wV) {
        super(s, label, wl, wV);
    }
}

abstract class TGStatement extends TGActionImpl {
    public TGStatement(A4CodeGen.State s, String label) {
        super(s, label);
    }
}

interface TGReader extends TGAction {
    public int getRV();
    public void setRV(int RV);
    public TGMemoryLocation getRl();
    public void setRl(TGMemoryLocation rl);
}

class TGReaderImpl extends TGActionImpl implements TGReader {
    private TGMemoryLocation rl;
    int rV;

    public TGReaderImpl(A4CodeGen.State s,
                    String label, TGMemoryLocation rl, int rV) {
        super(s, label); setRl(rl); this.rV = rV;
    }

    public int getRV() { return rV; }
    public void setRV(int rV) { this.rV = rV; }
    public TGMemoryLocation getRl() { return rl; }
    public void setRl(TGMemoryLocation rl) { this.rl = rl; }
}

class TGRemoteReadWrite extends TGWriterImpl implements TGRemoteAction, TGReader, TGWriter {
    private TGMemoryLocation rl;
    int rV;

    public TGRemoteReadWrite(A4CodeGen.State s, String label,
                             TGMemoryLocation rl, int rV,
                             TGMemoryLocation wl, int wV) {
        super(s, label, wl, wV);
    }

    public int getRV() { return rV; }
    public void setRV(int rV) { this.rV = rV; }
    public TGMemoryLocation getRl() { return rl; }
    public void setRl(TGMemoryLocation rl) { this.rl = rl; }
}

class TGRemoteRead extends TGReaderImpl implements TGRemoteAction {
    public TGRemoteRead(A4CodeGen.State s, String label,
                        TGMemoryLocation rl, int rV) {
        super(s, label, rl, rV);
    }
}

class TGExternRead extends TGReaderImpl implements TGExternAction {
    public TGExternRead(A4CodeGen.State s, String label,
                        TGMemoryLocation rl, int rV) {
        super(s, label, rl, rV);
    }
}

/* local read */
class TGRead extends TGReaderImpl implements TGExternAction {
    public TGRead(A4CodeGen.State s, String label,
                  TGMemoryLocation rl, int rV, TGRegister reg) {
        super(s, label, rl, rV);
        this.reg = reg;
    }

    TGRegister reg;

    @Override public String toString() {
        return String.format("load(atm) %s = %s",
                             reg==null ? "?" : reg.toString(), getRl());
    }
}

interface TGExternalAction extends TGAction {
}

interface TGRemoteAction extends TGAction {
}

interface TGExternAction extends TGExternalAction {
}

class TGRemoteWrite extends TGWriterImpl implements TGWriter, TGRemoteAction {
    public TGRemoteWrite(A4CodeGen.State s, String label,
                         TGMemoryLocation wl, int wV) {
        super(s, label, wl, wV);
    }
}

class TGExternWrite extends TGWriterImpl implements TGWriter, TGExternAction {
    public TGExternWrite(A4CodeGen.State s, String label,
                         TGMemoryLocation wl, int wV) {
        super(s, label, wl, wV);
    }
}

class TGRemoteGet extends TGStatement {
    Set<TGAction> actions;
    TGRemoteRead remoteRead;
    TGExternWrite externWrite;

    public TGRemoteGet(A4CodeGen.State s,
                       String label,
                       Set<TGAction> actions,
                       TGRemoteRead remoteRead,
                       TGExternWrite externWrite) {
        super(s, label);
        this.actions = actions;
        this.remoteRead = remoteRead;
        this.externWrite = externWrite;
    }

    @Override
    public String toString() {
        return String.format("%s = get(atm,atm)(%s,%s)", externWrite.getWl(), remoteRead.getRl(), remoteRead.getD().id);
    }
}

class TGRemotePut extends TGStatement {
    Set<TGAction> actions;
    TGExternRead externRead;
    TGRemoteWrite remoteWrite;

    public TGRemotePut(A4CodeGen.State s,
                       String label,
                       Set<TGAction> actions,
                       TGExternRead externRead,
                       TGRemoteWrite remoteWrite) {
        super(s, label);
        this.actions = actions;
        this.externRead = externRead;
        this.remoteWrite = remoteWrite;
    }

    @Override
    public String toString() {
        return String.format("put(atm,atm)(%s,%d,%s)", remoteWrite.getWl(), getD().id, externRead.getRl());
    }
}

class TGRemoteCompareAndSwap extends TGStatement {
    Set<TGAction> actions;
    TGExternRead externRead1;
    TGExternRead externRead2;
    TGExternWrite externWrite;
    TGRemoteReadWrite remoteReadWrite;

    public TGRemoteCompareAndSwap(A4CodeGen.State s,
                            String label,
                            Set<TGAction> actions,
                            TGExternRead externRead1,
                            TGExternRead externRead2,
                            TGExternWrite externWrite,
                            TGRemoteReadWrite remoteReadWrite) {
        super(s, label);
        this.actions = actions;
        this.externRead1 = externRead1;
        this.externRead2 = externRead2;
        this.externWrite = externWrite;
        this.remoteReadWrite = remoteReadWrite;
    }

    @Override
    public String toString() {
        return String.format("%s = cas(atm,atm,atm,atm)(%s,%d,%s,%s)", externWrite.getWl(), remoteReadWrite.getWl(), getD().id, externRead1.getRl(), externRead2.getRl());
    }
}

class TGRemoteGetAccumulate extends TGStatement {
    Set<TGAction> actions;
    TGExternWrite externWrite;
    TGExternRead externRead;
    TGRemoteReadWrite remoteReadWrite;

    public TGRemoteGetAccumulate(A4CodeGen.State s,
                                 String label,
                                 Set<TGAction> actions,
                                 TGExternWrite externWrite,
                                 TGExternRead externRead,
                                 TGRemoteReadWrite remoteReadWrite) {
        super(s, label);
        this.actions = actions;
        this.externWrite = externWrite;
        this.externRead = externRead;
        this.remoteReadWrite = remoteReadWrite;
    }

    @Override
    public String toString() {
        return String.format("%s = rga(atm,atm,atm)(%s,%d,%s)", externWrite.getWl(), remoteReadWrite.getWl(), getD().id, externRead.getRl());
    }
}

class TGFlush extends TGActionImpl {
    public TGFlush(A4CodeGen.State s, String label) {
        super(s, label);
    }

    @Override
    public String toString() {
        return String.format("flush (%d)", getD().id);
    }
}

public class A4CodeGen {
    class State {
        int memLocCount = 0;
        int registerCount = 0;
    }

    public final String MODULE;

    public String m(String mm) { return MODULE + mm; }

    public String ACTION_LABEL() { return m("Action"); }
    public String PO_FIELD_LABEL() { return "po"; }
    public String THR_LABEL() { return m("Thr"); }
    public String O_FIELD_LABEL() { return "o"; }
    public String D_FIELD_LABEL() { return "d"; }
    public String MEMORY_LOCATION_LABEL() { return m("MemoryLocation"); }
    public String REGISTER_LABEL() { return m("Register"); }
    public String HOST_FIELD_LABEL() { return "host"; }
    public String INITIAL_VALUE_LABEL() { return m("InitialValue"); }
    public String WRITER_LABEL() { return m("Writer"); }
    public String WL_FIELD_LABEL() { return "wl"; }
    public String WV_FIELD_LABEL() { return "wV"; }
    public String READER_LABEL() { return m("Reader"); }
    public String READ_LABEL() { return m("Read"); }
    public String AREAD_LABEL() { return m("ARead"); }
    public String NREAD_LABEL() { return m("NRead"); }
    public String RL_FIELD_LABEL() { return "rl"; }
    public String RV_FIELD_LABEL() { return "rV"; }
    public String REG_FIELD_LABEL() { return "reg"; }
    public String FLUSH_LABEL() { return m("Flush"); }
    public String RG_LABEL() { return m("Rg"); }
    public String RP_LABEL() { return m("Rp"); }
    public String RGA_LABEL() { return m("Rga"); }
    public String RCAS_LABEL() { return m("Rcas"); }
    public String STATEMENT_LABEL() { return m("Statement"); }
    public String ACTIONS_FIELD_LABEL() { return "actions"; }
    public String WITNESS_LABEL() { return m("Witness"); }

    A4Solution solution;
    State state;
    Map<String, TGAction> labelToActions = new HashMap<>();
    Map<String, TGMemoryLocation> labelToMemoryLocations =
        new HashMap<>();
    Map<String, TGRegister> labelToRegisters = new HashMap<>();
    Map<String, TGWriter> labelToWriters = new HashMap<>();
    Map<String, TGReader> labelToReaders = new HashMap<>();
    Map<String, TGFlush> labelToFlushes = new HashMap<>();
    Map<String, TGThread> labelToThreads = new HashMap<>();
    Map<TGThread, TGActionGraph> po = new HashMap<>();
    Map<TGAction, TGThread> actionToThread = new HashMap<>();

    // the next maps are currently write-only
    Map<String, TGRemoteGet> labelToRgs = new HashMap<>();
    Map<String, TGRemotePut> labelToRps = new HashMap<>();
    Map<String, TGRemoteGetAccumulate> labelToRgas = new HashMap<>();
    Map<String, TGRemoteCompareAndSwap> labelToRcass = new HashMap<>();

    TGRead witness;
    TGThread witnessThread;
    Map<TGThread, List<TGRead>> threadToLocalReads = new HashMap<>();

    public A4CodeGen(A4Solution solution, String m) {
        this.solution = solution;
        this.MODULE = m;
        this.state = new State();
    }

    Pattern stemPattern = Pattern.compile("(\\w*/)*[A-Z]?([A-Z]\\w*)(\\$.*)?");
    /* create a TGWriter of the appropriate subtype */
    /* NOTE: not fresh if RemoteReadWrite & already exists */
    TGWriter newTGWriter(A4CodeGen.State s,
                         String label, TGMemoryLocation wl, int wV) {
        Matcher m = stemPattern.matcher(label);
        if (m.matches()) {
            String stem = m.group(2);
            if (stem.equals("RemoteReadWrite")) {
                if (labelToReaders.containsKey(label)) {
                    TGRemoteReadWrite rw = (TGRemoteReadWrite)labelToReaders.get(label);
                    rw.setWl(wl); rw.setWV(wV);
                    return rw;
                }
                return new TGRemoteReadWrite(s, label, null, 0, wl, wV);
            }
            if (stem.equals("ExternWrite"))
                return new TGExternWrite(s, label, wl, wV);
            if (stem.equals("RemoteWrite"))
                return new TGRemoteWrite(s, label, wl, wV);
            if (stem.equals("InitialValue"))
                return new TGInitialValue(s, label, wl, wV);
            if (stem.equals("RemoteReadWrite"))
                return null;
        }
        return new TGWriterImpl(s, label, wl, wV, true);
    }

    /* create a TGReader of the appropriate subtype */
    /* NOTE: not fresh if RemoteReadWrite & already exists */
    TGReader newTGReader(A4CodeGen.State s,
                         String label,
                         TGMemoryLocation rl, int rV,
                         TGRegister reg) {
        Matcher m = stemPattern.matcher(label);
        if (m.matches()) {
            String stem = m.group(2);
            if (stem.equals("RemoteReadWrite")) {
                if (labelToWriters.containsKey(label)) {
                    TGRemoteReadWrite rw = (TGRemoteReadWrite)labelToWriters.get(label);
                    rw.setRl(rl); rw.rV = rV;
                    return rw;
                }
                return new TGRemoteReadWrite(s, label, rl, rV, null, 0);
            }
            if (stem.equals("RemoteRead"))
                return new TGRemoteRead(s, label, rl, rV);
            if (stem.equals("ExternRead"))
                return new TGExternRead(s, label, rl, rV);
            if (stem.equals("Read"))
                return new TGRead(s, label, rl, rV, null);
        }
        if (true) {
            System.out.println("error: " + label);
            System.exit(1);
        }
        return new TGReaderImpl(s, label, rl, rV);
    }

    void parseThreads() {
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(THR_LABEL())) {
                A4TupleSet threads = solution.eval(s);
                for (A4Tuple t : threads) {
                    String label = t.atom(0);
                    TGThread thread = new TGThread(state, label,
                                                   new ArrayList<>(),
                                                   new ArrayList<>(),
                                                   new ArrayList<>());
                    labelToThreads.put(label, thread);
                }
            }
        }
    }

    void parseMemoryLocations() {
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(MEMORY_LOCATION_LABEL())) {
                A4TupleSet mls = solution.eval(s);
                for (A4Tuple ml : mls) {
                    String label = ml.atom(0);
                    TGMemoryLocation memLoc = new TGMemoryLocation(state,
                                                                   label);
                    labelToMemoryLocations.put(label, memLoc);
                }

                for (Field f : s.getFields()) {
                    if (f.label.equals(HOST_FIELD_LABEL())) {
                        A4TupleSet thread4MemLoc = solution.eval(f);
                        for (A4Tuple t4m : thread4MemLoc) {
                            TGMemoryLocation m = labelToMemoryLocations.get(t4m.atom(0));
                            TGThread t = labelToThreads.get(t4m.atom(1));
                            t.memoryLocations.add(m);
                        }
                    }
                }
            }
        }
    }

    void parseRegisters() {
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(REGISTER_LABEL())) {
                A4TupleSet rs = solution.eval(s);
                for (A4Tuple rl : rs) {
                    String label = rl.atom(0);
                    TGRegister r = new TGRegister(state,
                                                  label);
                    labelToRegisters.put(label, r);
                }

                for (Field f : s.getFields()) {
                    if (f.label.equals(HOST_FIELD_LABEL())) {
                        A4TupleSet thread4Reg = solution.eval(f);
                        for (A4Tuple t4r : thread4Reg) {
                            TGRegister r = labelToRegisters.get(t4r.atom(0));
                            TGThread t = labelToThreads.get(t4r.atom(1));
                            t.registers.add(r);
                        }
                    }
                }
            }
        }
    }

    void parseWriters() {
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(WRITER_LABEL())) {
                A4TupleSet writers = solution.eval(s);
                for (A4Tuple writer : writers) {
                    String label = writer.atom(0);
                    TGWriter w = newTGWriter(state, label, null, 0);

                    labelToWriters.put(label, w);
                    labelToActions.put(label, w);
                }
                for (Field f : s.getFields()) {
                    if (f.label.equals(WV_FIELD_LABEL())) {
                        A4TupleSet val4Writer = solution.eval(f);
                        for (A4Tuple v4w : val4Writer) {
                            TGWriter w = labelToWriters.get(v4w.atom(0));
                            int v = Integer.parseInt(v4w.atom(1));
                            w.setWV(v);
                        }
                    }
                    if (f.label.equals(WL_FIELD_LABEL())) {
                        A4TupleSet loc4Writer = solution.eval(f);
                        for (A4Tuple l4w : loc4Writer) {
                            TGWriter w = labelToWriters.get(l4w.atom(0));
                            TGMemoryLocation l = labelToMemoryLocations.get(l4w.atom(1));
                            w.setWl(l);
                        }
                    }
                }
            }
        }
    }

    void parseReaders() {
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(READER_LABEL())) {
                A4TupleSet readers = solution.eval(s);
                for (A4Tuple reader : readers) {
                    String label = reader.atom(0);
                    TGReader r = newTGReader(state, label, null, 0, null);
                    labelToReaders.put(label, r);
                    labelToActions.put(label, r);
                }
                for (Field f : s.getFields()) {
                    if (f.label.equals(RV_FIELD_LABEL())) {
                        A4TupleSet val4Reader = solution.eval(f);
                        for (A4Tuple v4r : val4Reader) {
                            TGReader r = labelToReaders.get(v4r.atom(0));
                            int v = Integer.parseInt(v4r.atom(1));
                            r.setRV(v);
                        }
                    }
                    if (f.label.equals(RL_FIELD_LABEL())) {
                        A4TupleSet loc4Reader = solution.eval(f);
                        for (A4Tuple l4r : loc4Reader) {
                            TGReader r = labelToReaders.get(l4r.atom(0));
                            TGMemoryLocation l = labelToMemoryLocations.get(l4r.atom(1));
                            r.setRl(l);
                        }
                    }
                }
            }
        }

        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(READ_LABEL())) {
                for (Field f : s.getFields()) {
                    if (f.label.equals(REG_FIELD_LABEL())) {
                        A4TupleSet reg4Reader = solution.eval(f);
                        for (A4Tuple r4r : reg4Reader) {
                            TGRead r = (TGRead)labelToReaders.get(r4r.atom(0));
                            TGRegister l = labelToRegisters.get(r4r.atom(1));
                            r.reg = l;
                        }
                    }
                }
            }
        }
    }

    void parseFlushes() {
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(FLUSH_LABEL())) {
                A4TupleSet flushes = solution.eval(s);
                for (A4Tuple flush : flushes) {
                    String label = flush.atom(0);
                    TGFlush f = new TGFlush(state, label);
                    labelToFlushes.put(label, f);
                    labelToActions.put(label, f);
                }
                // parseOriginDestination handles assignment to threads
            }
        }
    }

    void parseInitialValues() {
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(INITIAL_VALUE_LABEL())) {
                A4TupleSet initialValues = solution.eval(s);
                for (A4Tuple iv : initialValues) {
                    TGWriter w = labelToWriters.get(iv.atom(0));
                    w.getWl().initialValue = w.getWV();
                }
            }
        }
    }

    void parseOriginDestinationThreads() {
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(ACTION_LABEL())) {
                A4TupleSet actions = solution.eval(s);
                for (A4Tuple a : actions) {
                    String label = a.atom(0);
                    if (labelToActions.get(label) == null) {
                        TGAction action = new TGActionImpl(state, label);
                        labelToActions.put(label, action);
                    }
                }
                for (Field f : s.getFields()) {
                    if (f.label.equals(O_FIELD_LABEL())) {
                        A4TupleSet thread4Action = solution.eval(f);
                        for (A4Tuple t4a : thread4Action) {
                            TGAction a = labelToActions.get(t4a.atom(0));
                            TGThread t = labelToThreads.get(t4a.atom(1));
                            t.actions.add(a);
                            actionToThread.put(a, t);
                        }
                    }
                    if (f.label.equals(D_FIELD_LABEL())) {
                        A4TupleSet thread4Action = solution.eval(f);
                        for (A4Tuple t4a : thread4Action) {
                            TGAction a = labelToActions.get(t4a.atom(0));
                            TGThread t = labelToThreads.get(t4a.atom(1));
                            a.setD(t);
                        }
                    }
                }
            }
        }
    }

    void parseProgramOrder() {
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(ACTION_LABEL())) {
                for (Field f : s.getFields()) {
                    if (f.label.equals(PO_FIELD_LABEL())) {
                        A4TupleSet pos = solution.eval(f);
                        for (A4Tuple po : pos) {
                            TGAction src = labelToActions.get(po.atom(0));
                            TGAction dest = labelToActions.get(po.atom(1));

                            src.getSuccs().add(dest);
                            dest.getPreds().add(src);
                        }
                    }
                }
            }
        }

        for (TGThread t : labelToThreads.values()) {
            po.put(t, new TGActionGraph(t.actions));
        }
    }

    void combineRop() {
        // rg has: label, actions, remote read, extern write
        // rp has: label, actions, extern read, remote write
        Map<String, Set<TGAction>> ropLabelToActions = new HashMap<>();

        // populate the 'actions' map based on Statement.actions
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(STATEMENT_LABEL())) {
                for (Field f : s.getFields()) {
                    if (f.label.equals(ACTIONS_FIELD_LABEL())) {
                        A4TupleSet actions = solution.eval(f);
                        for (A4Tuple action : actions) {
                            String rop = action.atom(0);
                            if (ropLabelToActions.get(rop) == null) {
                                ropLabelToActions.put(rop,
                                                      new HashSet<>());
                            }
                            Set<TGAction> rops = ropLabelToActions.get(rop);
                            rops.add(labelToActions.get(action.atom(1)));
                        }
                    }
                }
            }
        }

        // rg
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(RG_LABEL())) {
                A4TupleSet rgs = solution.eval(s);
                for (A4Tuple rgt : rgs) {
                    String label = rgt.atom(0);
                    Set<TGAction> actions = ropLabelToActions.get(label);
                    Set<TGAction> rgActions = new HashSet<>();

                    // find the remote read, extern write
                    TGExternWrite ew = null;
                    TGRemoteRead rr = null;
                    for (TGAction a : actions) {
                        if (a instanceof TGExternWrite) {
                            ew = (TGExternWrite)a;
                            rgActions.add(ew);
                        } else if (a instanceof TGRemoteRead) {
                            rr = (TGRemoteRead)a;
                            rgActions.add(rr);
                        }
                    }

                    assert rr != null && ew != null;
                    if (rr == null || ew == null) return;

                    assert rr.getSuccs().size() == 1 && rr.getSuccs().contains(ew);
                    assert ew.getPreds().size() >= 1 && ew.getPreds().contains(rr);
                    assert actionToThread.get(rr) != null;
                    assert actionToThread.get(rr).label == actionToThread.get(ew).label;
                    assert rr.getD() == ew.getD();

                    TGThread thr = actionToThread.get(rr);
                    TGActionGraph tag = po.get(thr);

                    TGRemoteGet rg = new TGRemoteGet(state, label, rgActions,
                                                     rr, ew);
                    rg.setD(rr.getD());

                    labelToRgs.put(label, rg);
                    labelToActions.put(label, rg);

                    tag.remove(rr);
                    tag.replace(ew, rg);
                }
            }
        }

        // rp
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(RP_LABEL())) {
                A4TupleSet rps = solution.eval(s);
                for (A4Tuple rpt : rps) {
                    String label = rpt.atom(0);
                    Set<TGAction> actions = ropLabelToActions.get(label);
                    Set<TGAction> rpActions = new HashSet<>();

                    // find the remote write, extern read
                    TGExternRead er = null;
                    TGRemoteWrite rw = null;
                    for (TGAction a : actions) {
                        if (a instanceof TGRemoteWrite) {
                            rw = (TGRemoteWrite)a;
                            rpActions.add(rw);
                        } else if (a instanceof TGExternRead) {
                            er = (TGExternRead)a;
                            rpActions.add(er);
                        }
                    }

                    assert rw != null && er != null;
                    if (rw == null || er == null) return;

                    assert er.getSuccs().size() == 1 && er.getSuccs().contains(rw);
                    assert rw.getPreds().size() >= 1 && rw.getPreds().contains(er);
                    assert rw.getD() == er.getD();
                    assert actionToThread.get(er) != null;
                    assert actionToThread.get(er).label == actionToThread.get(rw).label;

                    TGThread thr = actionToThread.get(er);
                    TGActionGraph tag = po.get(thr);

                    TGRemotePut rp = new TGRemotePut(state, label, rpActions,
                                                     er, rw);
                    rp.setD(rw.getD());

                    labelToRps.put(label, rp);
                    labelToActions.put(label, rp);

                    tag.remove(er);
                    tag.replace(rw, rp);
                }
            }
        }

        // rga
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(RGA_LABEL())) {
                A4TupleSet rgas = solution.eval(s);
                for (A4Tuple rgat : rgas) {
                    String label = rgat.atom(0);
                    Set<TGAction> actions = ropLabelToActions.get(label);
                    Set<TGAction> rgaActions = new HashSet<>();

                    // find the remote write, extern read
                    TGExternRead er = null;
                    TGRemoteReadWrite rrw = null;
                    TGExternWrite ew = null;
                    for (TGAction a : actions) {
                        if (a instanceof TGExternRead) {
                            er = (TGExternRead)a;
                            rgaActions.add(er);
                        } else if (a instanceof TGRemoteReadWrite) {
                            rrw = (TGRemoteReadWrite)a;
                            rgaActions.add(rrw);
                        } else if (a instanceof TGExternWrite) {
                            ew = (TGExternWrite)a;
                            rgaActions.add(ew);
                        }
                    }

                    assert er != null && rrw != null && ew != null;
                    if (er == null || rrw == null || ew == null) return;

                    TGThread thr = actionToThread.get(er);
                    TGActionGraph tag = po.get(thr);

                    TGRemoteGetAccumulate rga = new TGRemoteGetAccumulate(state, label, rgaActions,
                                                                          ew, er, rrw);
                    rga.setD(rrw.getD());

                    labelToRgas.put(label, rga);
                    labelToActions.put(label, rga);

                    tag.remove(er);
                    tag.remove(rrw);
                    tag.replace(ew, rga);
                }
            }
        }

        // rcas
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(RCAS_LABEL())) {
                A4TupleSet rcass = solution.eval(s);
                for (A4Tuple rcast : rcass) {
                    String label = rcast.atom(0);
                    Set<TGAction> actions = ropLabelToActions.get(label);
                    Set<TGAction> rcasActions = new HashSet<>();

                    // find the extern read (1 & 2), extern write, remote readwrite
                    List<TGExternRead> er = new ArrayList<>();
                    TGExternWrite ew = null;
                    TGRemoteReadWrite rrw = null;
                    for (TGAction a : actions) {
                        if (a instanceof TGExternWrite) {
                            ew = (TGExternWrite)a;
                            rcasActions.add(ew);
                        } else if (a instanceof TGExternRead) {
                            er.add((TGExternRead)a);
                            rcasActions.add(a);
                        } else if (a instanceof TGRemoteReadWrite) {
                            rrw = ((TGRemoteReadWrite)a);
                            rcasActions.add(rrw);
                        }
                    }

                    assert er.size() == 2 && ew != null && rrw != null;
                    if (!(er.size() == 2 && ew != null && rrw != null)) return;

                    TGThread thr = actionToThread.get(ew);
                    TGActionGraph tag = po.get(thr);

                    TGRemoteCompareAndSwap rcas = new TGRemoteCompareAndSwap
                        (state, label, rcasActions, er.get(0), er.get(1), ew, rrw);
                    rcas.setD(rrw.getD());

                    labelToRcass.put(label, rcas);
                    labelToActions.put(label, rcas);

                    tag.remove(er.get(0)); tag.remove(er.get(1));
                    tag.remove(ew);
                    tag.replace(rrw, rcas);
                }
            }
        }
    }

    void parseWitness() {
        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(WITNESS_LABEL())) {
                A4TupleSet ws = solution.eval(s);
                for (A4Tuple w : ws) {
                    witness = (TGRead)labelToReaders.get(w.atom(0));
                }
            }
        }
        if (witness != null) {
            witnessThread = actionToThread.get(witness);
        }
    }

    void parseLocalReads() {
        for (TGThread t : labelToThreads.values()) {
            threadToLocalReads.put(t, new ArrayList<TGRead>());
        }

        for (Sig s : solution.getAllReachableSigs()) {
            if (s.label.equals(AREAD_LABEL()) || s.label.equals(NREAD_LABEL())) {
                A4TupleSet rs = solution.eval(s);
                for (A4Tuple r : rs) {

                    TGRead read = (TGRead)labelToReaders.get(r.atom(0));
                    TGThread thread = actionToThread.get(read);

                    threadToLocalReads.get(thread).add(read);
                }
            }
        }
    }

    // also renumbers existing threads
    void eliminateEmptyThreads() {
        Set<TGThread> emptyThreads = new HashSet<>();
        for (TGThread t : labelToThreads.values()) {
            TGActionGraph tag = po.get(t);
            if (tag.actions.size() == 0)
                emptyThreads.add(t);
        }

        for (TGThread t : emptyThreads) {
            labelToThreads.remove(t.label);
        }

        int counter = 0;
        for (TGThread t : labelToThreads.values()) {
            t.id = counter++;
        }
    }

    void createModel() {
        parseThreads();
        parseMemoryLocations();
        parseRegisters();
        parseReaders();
        parseWriters();
        parseFlushes();
        parseOriginDestinationThreads();
        parseInitialValues();
        parseProgramOrder();
        parseWitness();
        parseLocalReads();
        combineRop();
        eliminateEmptyThreads();
    }

    void printThreads(PrintWriter w, int number) {
        int process_number = 1;
        for (TGThread t : labelToThreads.values()) {
            if (process_number > 1) w.println();
            w.format("process %d\n", t.id);
            process_number++;

            boolean first = true;
            if (!t.memoryLocations.isEmpty()) {
                StringBuilder memlocs = new StringBuilder("shared ");
                for (TGMemoryLocation ml : t.memoryLocations) {
                    if (first)
                        first = false;
                    else
                        memlocs.append(", ");

                    memlocs.append(ml);
                    if (ml.initialValue != null) {
                        memlocs.append(" = "); 
                        memlocs.append(ml.initialValue.toString());
                    }
                }
                memlocs.append(";");
                w.println(memlocs);
            }

            if (!t.registers.isEmpty()) {
                StringBuilder locals = new StringBuilder("local ");
                first = true;
                for (TGRegister r : t.registers) {
                    if (first)
                        first = false;
                    else
                        locals.append(", ");
                    locals.append(r);
                }
                locals.append(";");
                w.println(locals);
            }

            PseudoTopologicalOrderer<TGAction> pto = new PseudoTopologicalOrderer<>();
            for (TGAction a : pto.newList(po.get(t), false)) {
                if (a instanceof TGInitialValue)
                    continue;

                w.println(a + ";");
            }

            // check stage:
            //if (t == witnessThread && witness != null)
            //    w.printf("assert (!(%s == %s));\n", witness.reg, witness.rV);

            // also check the values of other local reads
            /*List<TGRead> localReads = threadToLocalReads.get(t);
            if ( ! localReads.isEmpty()) {
                TGRead lr = localReads.get(0);
                w.printf("assert (!((%s == %s)", lr.reg, lr.rV);

                for (int i = 1; i < localReads.size(); i++) {
                    lr = localReads.get(i);
                    w.printf(" && (%s == %s)", lr.reg, lr.rV);
                }

                w.printf("));\n");
            }*/



        }
        w.println("// ---\n");
    }

    public void emitCode(PrintWriter w, int number) {
        createModel();
        printThreads(w, number);
    }
}
